import numpy as np
from .distributions import uniform_generator, trunc_normal_sigma_generator,\
                           trunc_lognormal_sigma_generator
from functools import partial
from keyword import iskeyword


class Variable(object):
    '''
    Abstract class for keeping parameters of events in demographic model.
    Any new class for variable must be an instance of this class.

    New class should have :attr:`default_domain`, :attr:`default_rand_gen`
    class attributes and :func:`get_bounds`, :func:`get_possible_values`
    methods implemented.

    :cvar Variable.default_domain: default domain of the variable.
    :cvar Variable.default_rand_gen: default random generator of the variable.
                                     Is used in :func:`resample` method.

    :param name: unique name of the variable.
    :type name: str
    :param var_type: type of the variable (usually 'continuous'
                     or 'discrete').
    :type var_type: str
    :param domain: domain of the variable.
    :param rand_gen: random generator of the variable, should be a function
                     that takes domain as argument and returns sampled value.
    '''
    def __init__(self, name, var_type, domain, rand_gen):
        self.is_active = True
        self.name = name
        # TODO The following assert does not work for python 2
        assert(name.isidentifier() and not iskeyword(name))
        self.var_type = var_type
        self.domain = domain
        self.rand_gen = rand_gen

    def __copy__(self):
        return self

    def resample(self, *args, **kwargs):
        """
        Returns sampled value of the variable by calling `rand_gen` on
        the `domain`.

        :param `*args`: arguments to pass in `rand_gen`.
        :param `**kwargs`: kwargs to pass in `rand_gen`.
        """
        return self.rand_gen.__call__(self.domain, *args, **kwargs)

    def __str__(self):
        return "%s %s" % (self.__class__.__name__, self.name)

    def __repr__(self):
        return "%s %s" % (self.__class__.__name__, self.name)

    def get_bounds(self):
        """
        Returns bounds of the variable domain.
        """
        raise NotImplementedError

    def get_possible_values(self):
        """
        Returns list of all possible values of the variable.
        """
        raise NotImplementedError

    def correct_value(self, value):
        """
        Check that value is correct for this variable.
        """
        raise NotImplementedError


class ContinuousVariable(Variable):
    """
    Class of the continuous variable.

    :param domain: domain of the variable, if `None` then
                   :attr:`default_domain` will be taken.
    :param rand_gen: random generator for the variable, if `None` then
                     :attr:`default_rand_gen` will be taken.

    * :attr:`default_domain` = array([-inf, inf])

    * :attr:`default_rand_gen` = uniform distribution over domain.
    """
    default_domain = np.array([-np.inf, np.inf])
    default_rand_gen = uniform_generator

    def __init__(self, name, domain=None, rand_gen=None):
        if domain is None:
            domain = self.__class__.default_domain
        if rand_gen is None:
            rand_gen = self.__class__.default_rand_gen
        if domain[0] > domain[1]:
            raise ValueError("The lower bound of variable's domain must be "
                             "greater than upper bound. Got domain: "
                             f"{domain}.")
        super(ContinuousVariable, self).__init__(name, 'continuous',
                                                 domain, rand_gen)

    def get_bounds(self):
        """
        Returns the domain of the variable.
        """
        return self.domain

    def get_possible_values(self):
        """
        :raises AttributeError: it is impossible to get possible values\
        for continuous variable.
        """
        raise AttributeError("Impossible to produce a list of values for"
                             " continuous variable " + self.name)

    def correct_value(self, value):
        """
        Check that value is correct for this variable.
        """
        return value >= self.domain[0] and value <= self.domain[1]


class DiscreteVariable(Variable):
    """
    Class of the discrete variable.

    :param domain: domain of the variable, if `None` then
                   :attr:`default_domain` will be taken.
    :param rand_gen: random generator for the variable, if `None` then
                     :attr:`default_rand_gen` will be taken.

    """
    default_domain = np.array([])

    def default_rand_gen(a, size=None, replace=True, p=None):
        """
        See documentation of ``numpy.random.choice`` for more information.
        """
        return np.random.choice(a, size, replace, p)

    def __init__(self, name, domain=None, rand_gen=None):
        if domain is None:
            domain = self.__class__.default_domain
        if rand_gen is None:
            rand_gen = self.__class__.default_rand_gen
        super(DiscreteVariable, self).__init__(
                    name, 'discrete', domain, rand_gen)

    @property
    def domain(self):
        return self._domain

    @domain.setter
    def domain(self, domain):
        self._domain = np.array(domain, dtype=object)

    def get_bounds(self):
        """
        Returns bounds - minimum and maximum over domain of the variable.
        """
        return [min(self.domain), max(self.domain)]

    def get_possible_values(self):
        """
        Returns domain of the variable.
        """
        return self.domain

    def correct_value(self, value):
        """
        Check that value is correct for this variable.
        """
        return value in self.domain


class PopulationSizeVariable(ContinuousVariable):
    """
    Variable for keepeing size of population in demographic model.

    * :attr:`default_domain` = array([1e-2, 100])

    * :attr:`default_rand_gen` = truncated log normal distribution over\
        domain with mean equal to 1.

    :note: Values are assumed to be in genetic units.
    """
    default_domain = np.array([1e-2, 100])
    default_rand_gen = trunc_lognormal_sigma_generator

    @staticmethod
    def translate_units(value, Nanc):
        return type(Nanc)(value * Nanc)


def migration_generator(domain):
    """
    Generates random value of the migration. With probability of 0.5 generates
    zero, otherwise runs
    :func:`gadma.utils.distributions.trunc_normal_sigma_generator`.
    """
    if 0 == domain[0] and np.random.choice([False, True]):
        return 0
    return trunc_normal_sigma_generator(domain)


class MigrationVariable(ContinuousVariable):
    """
    Variable for keepeing migration parameter of the demographic model.

    * :attr:`default_domain` = array([0, 10])

    * :attr:`default_rand_gen` = truncated log normal distribution over\
      domain with mean equal to 1.

    :note: Values are assumed to be in genetic units.
    """
    default_domain = np.array([0, 10])
    default_rand_gen = migration_generator

    @staticmethod
    def translate_units(value, Nanc):
        return value / (2 * Nanc)


class TimeVariable(ContinuousVariable):
    """
    Variable for keepeing time parameter of the demographic model.

    * :attr:`default_domain` = array([0, 5])

    * :attr:`default_rand_gen` = truncated log normal distribution over\
        domain with mean equal to 1.

    :note: Values are assumed to be in genetic units.
    """
    default_domain = np.array([1e-15, 5])
    default_rand_gen = trunc_normal_sigma_generator

    @staticmethod
    def translate_units(value, Nanc):
        return type(Nanc)(value * (2 * Nanc))


class SelectionVariable(ContinuousVariable):
    """
    Variable for keepeing selection parameter of the demographic model.

    * :attr:`default_domain` = array([0, 10])

    * :attr:`default_rand_gen` = truncated log normal distribution over\
        domain with mean equal to 1.

    :note: Values are assumed to be in genetic units.
    """
    default_domain = np.array([1e-15, 10])
    default_rand_gen = trunc_normal_sigma_generator

    @staticmethod
    def translate_units(value, Nanc):
        return value / (2 * Nanc)


class FractionVariable(ContinuousVariable):
    """
    Variable for keepeing fraction parameter of the demographic model.

    * :attr:`default_domain` = array([0, 1])

    * :attr:`default_rand_gen` = random uniform distribution over domain.
    """
    default_domain = np.array([1e-3, 1-1e-3])
    default_rand_gen = uniform_generator

    @staticmethod
    def translate_units(value, Nanc):
        return value


class Dynamic(object):
    """
    Abstract class for Dynamic value.
    New class should be instance of this class, should have :attr:`format_str`
    attribute and implement :func:`_inner_func` (staticmethod),
    :func:`__str__` and :func:`func_str` methods.

    :cvar Dynamic.format_str: format string for string representation of the
                              dynamic.
    """
    format_str = ''

    @staticmethod
    def _inner_func(y1, y2, x_diff):
        """
        Inner function of the dynamic.

        :param y1: the initial value of the dynamic.
        :param y2: the final value of the dynamic.
        :param x_diff: the x-difference between initial and final values.
        """
        raise NotImplementedError

    def __str__(self):
        """
        Returns string representation of the dynamic.
        """
        raise NotImplementedError

    @classmethod
    def func_str(cls, y1, y2, x_diff):
        """
        Returns string representation of the dynamic via formating
        :attr:`format_str` with argumets.
        """
        return cls.format_str.format(str(y1), str(y2), str(x_diff))


def linear(y1, y2, x_diff, x):
    y = y1 + (y2 - y1) * (x / x_diff)
    return y


def exponent(y1, y2, x_diff, x):
    y = y1 * (y2 / y1) ** (x / x_diff)
    return y


def constant(y1, y2, x_diff, x):
    return y2


class Exp(Dynamic):
    """
    Exponential dynamic.

    * :attr:`format_str` = "lambda t: {0} * ({1} / {0}) ** (t / {2})"
    """
    format_str = "lambda t: {0} * ({1} / {0}) ** (t / {2})"

    @staticmethod
    def _inner_func(y1, y2, x_diff):
        """
        Returns lambda function of x: y1 * (y2 / y1) ** (x / x_diff).
        """
        return partial(exponent, y1, y2, x_diff)

    def __str__(self):
        """
        Returns "Exp".
        """
        return "Exp"


class Lin(Dynamic):
    """
    Linear dynamic.

    * :attr:`format_str` = "lambda t: {0} + ({1} - {0}) * (t / {2})"
    """
    format_str = "lambda t: {0} + ({1} - {0}) * (t / {2})"

    @staticmethod
    def _inner_func(y1, y2, x_diff):
        """
        Returns lambda function of x: y1 + (y2 - y1) * (x / x_diff).
        """
        return partial(linear, y1, y2, x_diff)

    def __str__(self):
        """
        Returns "Lin".
        """
        return "Lin"


class Sud(Dynamic):
    """
    Sudden (constant) dynamic.

    * :attr:`format_str` = "{1}"
    """
    format_str = '{1}'

    @staticmethod
    def _inner_func(y1, y2, x_diff):
        """
        Returns y2.
        """
        return partial(constant, y1, y2, x_diff)

    def __str__(self):
        """
        Returns "Sud".
        """
        return "Sud"


def dynamic_generator(domain):
    p = []
    for x in domain:
        if x == 'Sud' or x == 0:
            p.append(len(domain) + 1)
        elif x == 'Lin' or x == 1:
            p.append(len(domain) - 1)
        else:
            p.append(len(domain))
    p = np.array(p) / np.sum(p)
    return np.random.choice(domain, p=p)


class DynamicVariable(DiscreteVariable):
    """
    Variable for keepeing selection parameter of the demographic model.

    * :attr:`default_domain` = array([\
        :class:`gadma.utils.variables.Exp`,\
        :class:`gadma.utils.variables.Lin`,\
        :class:`gadma.utils.variables.Sud`])

    * :attr:`default_rand_gen` = random choice over domain.
    """

    _help_dict = {'Sud': Sud, 'Lin': Lin, 'Exp': Exp, 0: Sud, 1: Lin, 2: Exp}
    default_domain = np.array(['Sud', 'Lin', 'Exp'])
    default_rand_gen = dynamic_generator

    def __init__(self, name, domain=None, rand_gen=None):
        if domain is None:
            domain = self.__class__.default_domain
        if not all(dom in self.__class__._help_dict for dom in domain):
            raise ValueError("Domain of DynamicVariable must be a subset "
                             "of the following general domain: "
                             f"{self._help_dict.keys()}")
        super(DynamicVariable, self).__init__(name, domain, rand_gen)

    @staticmethod
    def get_func_from_value(value):
        """
        Returns :func:`gadma.utils.variables.Dynamic._inner_func` function
        from the value.

        :param value: value of the variable.
        :type value: :class:`gadma.utils.variables.Dynamic`
        """
        if value not in DynamicVariable._help_dict:
            raise ValueError(f"Value {value} not in domain:"
                             f"{DynamicVariable._help_dict.keys()}.")
        return DynamicVariable._help_dict[value]._inner_func

    def get_bounds(self):
        """
        :raises AttributeError: Dynamic variable has incomparative values.
        """
        raise AttributeError("DynamicVariable domain has incomparative values")

    @staticmethod
    def translate_units(value, Nanc):
        return value



import numpy as np
from ..utils import Variable, VariablePool, float_repr


class Model(object):
    """
    Abstract class of model. Contains variables of class :class:`Variable`.

    :param raise_excep: if `True` then raises exception when something except
                        :class:`Variable` is added as variable.
    :type raise_excep: bool
    """
    def __init__(self, raise_excep=False):
        self.is_fixed = []
        self.fixed_values = {}
        self._variables = VariablePool()
        self.raise_excep = raise_excep

    @property
    def variables(self):
        return [var for var, fixed in zip(self._variables, self.is_fixed)
                if not fixed]

    @variables.setter
    def variables(self, new_variables):
        self._variables = new_variables
        self.is_fixed = [False for var in new_variables]

    def add_variable(self, variable):
        """
        Adds one variable to the model.

        :param variable: variable to add.
        :type variable: :class:`Variable`
        """
        if not isinstance(variable, (Variable, Model)):
            if self.raise_excep:
                raise ValueError("Only instances of class Variable could be"
                                 " added to the model as variables.")
        else:
            if isinstance(variable, Model):
                for var in variable.variables:
                    self.add_variable(var)
                return
            if variable not in self._variables:
                self._variables.append(variable)
                self.is_fixed.append(False)

    def add_variables(self, variables):
        """
        Adds several variables to the model.

        :param variables: variables to add.
        :type variables: list
        """
        for variable in np.array(variables).flatten():
            self.add_variable(variable)

    def get_variable(self, name):
        """
        Finds the variable of the model by its name.

        :param name: name of the model's variable.
        :type name: str

        :raises ValueError: if there is no variable with that name.
        """
        for var in self.variables:
            if var.name == name:
                return var

    def fix_variable(self, variable, value):
        if variable not in self.variables:
            raise ValueError(f"There is no such unfixed variable {variable} "
                             "in the model.")
        self.is_fixed[self._variables.index(variable)] = True
        self.fixed_values[variable] = value

    def unfix_variable(self, variable):
        if variable not in self.fixed_values:
            raise ValueError(f"There is no such fixed variable {variable} "
                             "in the model.")
        ind = self._variables.index(variable)
        self.is_fixed[ind] = False
        del self.fixed_values[variable]

    def unfix_if_fixed(self, variable):
        try:
            self.unfix_variable(variable)
        except ValueError:
            pass

    def var2value(self, values):
        """
        Returns dictionary {variable: value}.

        :param values: List or dict {var_name: value} of values.
        """
        if len(self.variables) == 0:
            return {}
        if isinstance(values, list) or isinstance(values, np.ndarray):
            ret_dict = {var: value for var, value in zip(self.variables,
                                                         values)}
        elif isinstance(values, dict):
            ret_dict = {}
            for key in values:
                if isinstance(key, str):
                    var = self.get_variable(key)
                    if var is not None:
                        ret_dict[var] = values[key]
                elif isinstance(key, Variable):
                    if key in self.variables:
                        ret_dict[key] = values[key]
                    else:
                        var = self.get_variable(key.name)
                        if var is not None:
                            ret_dict[var] = values[key]
        else:
            raise TypeError("Values are either not list nor dict.")

        assert len(ret_dict) == len(self.variables)
        return {**ret_dict, **self.fixed_values}

    def string_repr(self, values):
        """
        Returns string representation of variables and values.

        :param values: Values of the variables in model.
        """
        strings = []
        var2value = self.var2value(values)
        for var in self.variables:
            strings.append(f"{var.name}={var2value[var]}")
        return ", ".join(strings)

    def _arg_val_repr(self, arg, values):
        """
        Returns list of arg and its value string representations
        """
        from .variables_combinations import BinaryOperation
        if isinstance(arg, Variable):
            val = self.var2value(values)[arg]
            arg_repr = f"{arg.name}"
        elif isinstance(arg, BinaryOperation):
            var2value = self.var2value(values)
            val = arg.get_value([var2value[var] for var in arg.variables])
            arg_repr = f"{arg.name}"
        else:
            val = arg
            arg_repr = ""
        if isinstance(val, float):
            val_repr = float_repr(val, precision=3)
        else:
            val_repr = str(val)
        return arg_repr, val_repr



from .. import matplotlib, moments, dadi, Image
from .. import matplotlib_available, PIL_available, moments_available
from ..models import EpochDemographicModel, Split
from ..engines import get_engine, all_engines
from ..utils import bcolors, WeightedMetaArray

import warnings
import os
import warnings
import io
from datetime import datetime
import copy
import numpy as np


def draw_plots_to_file(x, engine, settings, filename, fig_title):
    """
    Draws plots of data (SFS) and model from ``engine`` with parameters
    ``x``.

    :param x: Values of the parameters in model.
    :type x: list or dict
    :param engine: Engine with specified model and data.
    :type engine: :class:`gadma.engines.engine.Engine`
    :param filename: File name to save picture.
    :type filename: str
    :param fig_title: Title of the schematic model plot.
    :type fig_title: str

    :note: print warnings if schematic model plot was not drawn.
    """
    if not matplotlib_available:
        raise ValueError("Matplotlib is required to draw models.")
        return
    warnings.filterwarnings(
            'ignore', category=matplotlib.cbook.MatplotlibDeprecationWarning)

    # 0. Get place of dot in filename - where file extention starts
    pos = filename.rfind('.')
    if pos == -1:
        pos = len(filename)

    # 1. Draw sfs
    # 1.1 Set file or buffer to save plot
    if PIL_available:  # then we will concatenate plots later
        if not moments_available:  # then we won't draw model plot
            save_file_sfs = filename
        else:  # future concatenation
            save_file_sfs = io.BytesIO()
    else:
        save_file_sfs = filename[:pos] + '_sfs' + filename[pos:]
    # 1.2 Draw plot to save_file
    engine.draw_sfs_plots(x, *settings.get_engine_args(),
                          save_file=save_file_sfs, vmin=settings.vmin)

    # 2 Draw schematic model plot
    # 2.0 Check that moments is available, it not we return
    if not moments_available:
        raise ValueError("Moments is required to draw schematic model plots.")

    # 2.1 Set file or buffer to save plot
    if PIL_available:
        save_file_model = io.BytesIO()
    else:
        save_file_model = filename[:pos] + '_model' + filename[pos:]
    # 2.2 Create all arguments for drawing function
    nref = engine.get_N_ancestral(x, *settings.get_engine_args())
    gen_time = settings.time_for_generation
    if gen_time is None:
        gen_time = 1.0
        gen_time_units = "Generations"
    else:
        gen_time *= settings.const_of_time_in_drawing
        gen_time_units = settings.units_of_time_in_drawing
    # 2.3 Draw model plot with moments engine
    # We use try except to be carefull
    try:
        engine.draw_schematic_model_plot(x, save_file_model, fig_title, nref,
                                         gen_time, gen_time_units)
    except Exception as e:
        save_file_sfs.seek(0)
        with open(filename, 'wb') as fl:
            fl.write(save_file_sfs.read())
        raise e

    # 3. Concatenate plots if PIL is available
    if PIL_available:
        save_file_sfs.seek(0)
        save_file_model.seek(0)

        img1 = Image.open(save_file_model)
        img2 = Image.open(save_file_sfs)

        if img2.size[1] < img1.size[1]:
            img2 = img2.resize(
                (int(img1.size[1] * img2.size[0] / img2.size[1]),
                 img1.size[1]))

        width = img1.size[0] + img2.size[0]
        height = max(img1.size[1], img2.size[1])

        new_img = Image.new('RGB', (width, height))

        new_img.paste(img1, (0, 0))
        new_img.paste(img2, (img1.size[0], 0))

        new_img.save(filename)


def generate_code_to_file(x, engine, settings, filename):
    """
    Generates code of demographic model to file. Settings are required to get
    ``engine`` arguments in :func:`evaluation` function.

    :param x: Values of the parameters in model.
    :type x: list or dict
    :param engine: Engine with specified model and data.
    :type engine: :class:`gadma.engines.engine.Engine`
    :param settings: Settings of the run.
    :type settings: :class:`gadma.cli.settings_storage.SettingsStorage`
    :param filename: File name to save picture.
    :type filename: str
    """
    pos = filename.rfind('.')
    if pos == -1 or filename[pos:] != '.py':
        pos = len(filename)
    prefix = filename[:pos]

    nref = engine.get_N_ancestral(x, *settings.get_engine_args())
    gen_time = None
    if settings.time_for_generation is not None:
        gen_time = settings.time_for_generation *\
                   settings.const_of_time_in_drawing
    gen_time_units = settings.units_of_time_in_drawing
    # Generate code
    if isinstance(engine.model, EpochDemographicModel):
        engines = all_engines()
    else:
        engines = [copy.deepcopy(engine)]
    for other_engine in engines:
        save_file = prefix + f"_{other_engine.id}_code.py"
        other_engine.set_data(engine.data)
        other_engine.data_holder = copy.deepcopy(engine.data_holder)
        other_engine.set_model(engine.model)
        args = settings.get_engine_args(other_engine.id)
        other_engine.generate_code(x, save_file, *args, nref, gen_time,
                                   gen_time_units)


def print_runs_summary(start_time, shared_dict, settings):
    """
    Prints best demographic model by logLL among all processes.

    :param start_time: Time when equation was started.
    :type start_time: float
    :param shared_dict: Dictionary to share information between processes.
    :type shared_dict: :class:`gadma.core.shared_dict.SharedDict`
    :param settings: Settings of run.
    :type settings: :class:`gadma.cli.settings_storage.SettingsStorage`
    """
    s = (datetime.now() - start_time).total_seconds()
    time_str = f"\n[{int(s//3600):03}:{int(s%3600//60):02}:{int(s%60):02}]"
    print(time_str)
    metric_names = shared_dict.get_available_groups()
    if len(metric_names) == 0:
        print("No models yet")
    for best_by in metric_names:
        models = shared_dict.get_models_in_group(best_by, align_y_dict=True)
        local_metrics = models[0][1][2].keys()
        sorted_models = models
        metrics = local_metrics

        print(f"All best by {best_by} models")
        print("Number", *metrics, "Model", sep='\t')
        for model in sorted_models:
            index, info = model
            engine, x, y_vals = info
            # Get theta and N ancestral
            theta = engine.get_theta(x, *settings.get_engine_args())
            Nanc = engine.get_N_ancestral_from_theta(theta)
            addit_str = f"(theta = {theta: .2f})"
            if Nanc is not None:
                Nanc = int(Nanc)
                if settings.relative_parameters:
                    addit_str += f" (Nanc = {Nanc: d})"
                    model_str = engine.model.as_custom_string(x)
                else:
                    model_str = f" [Nanc = {Nanc: d}] "
#                    Tg = settings.time_for_generation or 1.0
                    x_translated = engine.model.translate_units(x, Nanc)
                    model_str += engine.model.as_custom_string(x_translated)
            else:
                model_str = engine.model.as_custom_string(x)
            if hasattr(x, "metadata"):
                model_str += f"\t{x.metadata}"
            # Begin to print
            metric_vals = []
            for metr in metrics:
                if metr not in y_vals:
                    metric_vals.append("None")
                else:
                    if isinstance(y_vals[metr], tuple):
                        if y_vals[metr][0] is None:
                            val_str = "None"
                        else:
                            val_str = f"{y_vals[metr][0]:.2f} "\
                                      f"(eps={y_vals[metr][1]:.1e})"
                    else:
                        val_str = f"{y_vals[metr]:.2f}"
                    metric_vals.append(val_str)
            print(f"Run {index}", *metric_vals, model_str,
                  addit_str, sep='\t')
            fig_title = f"Best by {best_by} model. "
            ind = 0
            for metr in metrics:
                if metr not in y_vals:
                    continue
                val_str = metric_vals[ind]
                ind += 1
                fig_title += f"{metr}: {val_str}"
        # Draw and generate code for best model
        _, (engine, x, y_vals) = sorted_models[0]
        prefix = (settings.BASE_OUTPUT_DIR_PREFIX +
                  settings.LONG_NAME_2_SHORT.get(best_by.lower(),
                                                 best_by.lower()))
        out_dir = settings.output_directory
        save_plot_file = os.path.join(out_dir, prefix + "_model.png")
        save_code_file = os.path.join(out_dir, prefix + "_model.py")
        drawn = True
        gener = True
        try:
            draw_plots_to_file(x, engine, settings, save_plot_file, fig_title)
        except Exception as e:
            drawn = False
            print(f"{bcolors.FAIL}Run {index}: failed to draw model due to "
                  f"the following exception: {e}{bcolors.ENDC}")
        try:
            generate_code_to_file(x, engine, settings, save_code_file)
        except Exception as e:
            gener = False
            print(f"{bcolors.FAIL}Run {index}: failed to generate code due to"
                  f" the following exception: {e}{bcolors.ENDC}")
        if drawn and gener:
            print("\nYou can find picture and python code of best model in "
                  "the output directory.\n")
        elif drawn:
            print("\nYou can find picture of best model in the output "
                  "directory.\n")
        elif gener:
            print("\nYou can find python code of best model in the output "
                  "directory.\n")



from ..cli import arg_parser
from ..utils import StdAndFileLogger, bcolors

from .draw_and_generate_code import print_runs_summary
from .core_run import CoreRun
from .shared_dict import SharedDictForCoreRun

from functools import partial
import numpy as np
import os
import sys

from datetime import datetime
import operator
import multiprocessing
import signal
from multiprocessing import Manager, Pool
import math
import warnings
import time
import traceback


SUPPORT_STRING = "\nIn case of any questions or problems, "\
                 "please contact: ekaterina.e.noskova@gmail.com\n"


def job(index, shared_dict, settings):
    """
    Function of one parallel run of GADMA. Creates
    :class:`gadma.core.core_run.CoreRun` object and call its :meth:`run`
    method.

    :param index: Index of the run.
    :type index: int
    :param shared_dict: Shared dict between all runs.
    :type shared_dict: :class:`gadma.core.shared_dict.SharedDictForCoreRun`
    :param settings: Settings of the run.
    :type settings: :class:`gadma.cli.settings_storage.SettingsStorage`
    """
    try:
        obj = CoreRun(index, shared_dict, settings)
        obj.run(settings.get_optimizers_init_kwargs())
    except Exception as e:
        print(f"{bcolors.FAIL}Run {index} failed due to following exception:"
              f"{bcolors.ENDC}", file=sys.stderr)
        print(traceback.format_exc(), file=sys.stderr)
        raise e


def main():
    """
    Main function that is called from command line. Creates parallel runs of
    GADMA and holds base pool of processes. Prints progress periodically for
    each run, saves plots and pictures of best model and generates code.
    """
    settings_storage, args = arg_parser.get_settings()

    # Form output directory
    log_file = os.path.join(settings_storage.output_directory, 'GADMA.log')
    params_file = os.path.join(settings_storage.output_directory,
                               'params_file')
    extra_params_file = os.path.join(settings_storage.output_directory,
                                     'extra_params_file')

    # Change output stream both to stdout and log file
    saved_stdout = sys.stdout
    saved_stderr = sys.stderr
    sys.stdout = StdAndFileLogger(log_file, settings_storage.silence)
    sys.stderr = StdAndFileLogger(log_file)

    print(f"{bcolors.OKGREEN}--Successful arguments parsing--{bcolors.ENDC}")

    # Data reading
    print("\nData reading")
    data = settings_storage.read_data()
    print(f"Number of populations: {settings_storage.number_of_populations}")
    print(f"Projections: {settings_storage.projections}")
    print(f"Population labels: {settings_storage.population_labels}")
    print(f"Outgroup: {settings_storage.outgroup}")
    print(f"{bcolors.OKGREEN}--Successful data reading--{bcolors.ENDC}")

    if settings_storage.directory_with_bootstrap is not None:
        print("\nBootstrap data reading")
        boot_data = settings_storage.read_bootstrap_data()
        print(f"Number of files found: "
              f"{len(settings_storage.bootstrap_data)}")
        print(f"{bcolors.OKGREEN}--Successful bootstrap data reading--"
              f"{bcolors.ENDC}")

    # Save parameters
    settings_storage.to_files(params_file, extra_params_file)
    if not args.test:
        print(f"\nParameters of launch are saved in output directory: "
              f"{params_file}")
        print(f"All output is saved in output directory: {log_file}")

    print(f"{bcolors.OKBLUE}--Start pipeline--{bcolors.ENDC}")

    # Create shared dictionary
    shared_dict = SharedDictForCoreRun()

    # Start pool of processes
    start_time = datetime.now()

    # For debug
#    shared_dict = SharedDictForCoreRun(multiprocessing=False)
#    job(0, shared_dict, settings_storage)
#    os._exit(0)

    pool = Pool(processes=settings_storage.number_of_processes)

    results = []
    for proc_args in [(i + 1, shared_dict, settings_storage)
                      for i in range(settings_storage.number_of_repeats)]:
        results.append(pool.apply_async(job, proc_args))

    pool.close()

    check_time = time.time()
    time_diff = 60 * settings_storage.time_to_print_summary
    time_bias = 0
    get_time = 10
    while True:
        time.sleep(get_time - time_bias)
        time_bias = 0
        all_finished = True
        for r in results:
            try:
                r.get(0)
            except multiprocessing.TimeoutError as e:
                all_finished = False
            except Exception as e:
                pool.terminate()
                print(f"{bcolors.FAIL}Main run failed due to following "
                      f"exception:{bcolors.ENDC}", file=sys.stderr)
                print(traceback.format_exc(), file=sys.stderr)
                print(SUPPORT_STRING)
                sys.stdout = saved_stdout
                sys.stderr = saved_stderr
                os._exit(1)
        if all_finished:
            break
        if (time.time() - check_time) >= time_diff:
            check_time = time.time()
            print_runs_summary(start_time, shared_dict, settings_storage)
            time_bias = time.time() - check_time
            time_bias %= get_time

    pool.join()
    print_runs_summary(start_time, shared_dict, settings_storage)

    print('\n--Finish pipeline--\n')
    if args.test:
        print('--Test passed correctly--')
    if settings_storage.theta0 is None:
        print("\nYou didn't specify theta at the beginning. If you want "
              "change it and rescale parameters, please see tutorial.\n")
#        if (params.resume_dir is not None and
#                (params.initial_structure != params.final_structure).any()):
#            print('\nYou have resumed from another launch. Please, check '
#                  'best AIC model, as information about it was lost.\n')

    print('Thank you for using GADMA!')
    print(SUPPORT_STRING)

    sys.stdout = saved_stdout
    sys.stderr = saved_stderr


if __name__ == '__main__':
    main()



import tempfile
from ..version import __version__
from .settings_storage import HOME_DIR
from . import SettingsStorage
from ..core import SUPPORT_STRING
from ..utils import ensure_dir_existence

import warnings
import copy
import argparse
import sys
import os

TEST_SETTINGS = os.path.join(HOME_DIR, "test_settings")


def version():
    '''
    Returns string with current version.
    '''
    return "GADMA version " + str(
        __version__
    ) + "\tby Ekaterina Noskova (ekaterina.e.noskova@gmail.com)" + "\n"


def usage():
    '''
    Returns usage of tool.
    '''
    return version() + "" \
        "Usage: \n\tgadma\t-p/--params <params_file>\n"\
        "\t\t-e/--extra <extra_params_file>\n"\
        "\n\n"\
        "Instead/With -p/--params and -e/--extra option you can set:\n"\
        "\t-o/--output <output_dir>\toutput directory.\n"\
        "\t-i/--input <in.fs>/<in.txt>\tinput file with AFS or in dadi "\
        "format.\n"\
        "\t--resume <resume_dir>\t\tresume another launch from "\
        "<resume_dir>.\n"\
        "\t--only_models\t\t\tflag to take models only from another launch "\
        "(--resume option).\n\n"\
        "\t-h/--help\t\tshow this help message and exit.\n"\
        "\t-v/--version\t\tshow version and exit.\n"\
        "\t--test\t\t\trun test case.\n" + SUPPORT_STRING


class ArgParser(argparse.ArgumentParser):
    """
    Overrided class for argument parser.
    """
    def format_help(self):
        """
        Returns usage by calling :func:`usage`.
        """
        return usage()


def test_args():
    '''
    Put default args for test case.
    '''
    # Create storage from test settings
    settings_storage = SettingsStorage.from_file(TEST_SETTINGS)
    # Input test file
    curent_dir = os.path.dirname(os.path.abspath(__file__))
    settings_storage.input_file = os.path.join(curent_dir,
                                               "../../fs_examples/test.fs")
    # There is no output_directory, we put it to temporary directory
    settings_storage.output_directory = tempfile.mkdtemp("gadma_test_dir")
    # And put path to input file
    settings_storage.input_file = os.path.join(HOME_DIR,
                                               settings_storage.input_file)
    return settings_storage


def get_settings():
    '''
    Parse args from command line and store them in options_storage.

    :returns: tuple of parsed arguments and settings storage.
    '''
    usage = str(sys.argv[0]) + " -p <params_file>"

    # Create arguments parser
    parser = ArgParser(add_help=False)
    parser.add_argument('-p', '--params', metavar="<params_file>",
                        required=False, default=None)
    parser.add_argument('-e', '--extra', metavar="<extra_params_file>",
                        required=False, default=None)
    parser.add_argument('--resume', metavar="<dir>",
                        required=False, default=None)
    parser.add_argument('--only_models', action='store_true')
    parser.add_argument('-o', '--output', metavar="<dir>",
                        required=False, default=None)
    parser.add_argument('-i', '--input',
                        metavar="<filename.fs>/<filename.txt>",
                        required=False, default=None)
    parser.add_argument('-v', '--version', action='version',
                        version=version(), default=argparse.SUPPRESS)
    parser.add_argument('--test', action='store_true')
    parser.add_argument('-h', '--help', action='help',
                        default=argparse.SUPPRESS)

    # If not enough arguments
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    # 1. Parse arguments
    args = parser.parse_args()

    # 2. Create Settings storage
    if args.test:
        print("--Running test case--")
        settings_storage = test_args()
        # We return as we do not need to check
        return settings_storage, args
    else:
        settings_storage = SettingsStorage.from_file(args.params, args.extra)

    # Check for resume
    if args.resume is not None:
        args.resume = os.path.abspath(os.path.expanduser(args.resume))
        if (settings_storage.resume_from is not None and
                settings_storage.resume_from != args.resume):
            raise ValueError(f"Resume directory in parameters file "
                             f"({settings_storage.resume_from}) doesn't "
                             f"match to one from the --resume option "
                             f"({args.resume})")
        settings_storage.resume_from = args.resume

    if args.only_models:
        settings_storage.only_models = True

    if settings_storage.resume_from is not None:
        old_params_file = os.path.join(settings_storage.resume_from,
                                       'params_file')
        old_extra_file = os.path.join(settings_storage.resume_from,
                                      'extra_params_file')
        if not os.path.exists(old_params_file):
            warnings.warn("There is no saved params file in resume directory. "
                          "Resume option will be ignored.")
            settings_storage.resume_from = None
            settings_storage.only_models = False
        else:
            if not os.path.exists(old_extra_file):
                old_extra_file = None
            resume_from_settings = SettingsStorage.from_file(
                old_params_file, old_extra_file)
            if (settings_storage.output_directory is None and
                    args.output is None):
                resume_from_settings.output_directory += "_resumed"
            else:
                resume_from_settings.output_directory = None
            settings_storage = copy.copy(resume_from_settings)
            settings_storage = settings_storage.update_from_file(args.params,
                                                                 args.extra)
            settings_storage.only_models = False
            if args.resume:
                settings_storage.resume_from = args.resume

    if args.output is not None:
        if (settings_storage.output_directory is not None and
                settings_storage.output_directory != args.output):
            warnings.warn("Output directory in parameters file doesn't match "
                          "to one from the -o/--output option. "
                          "The last is taken.")
        settings_storage.output_directory = args.output

    if args.input is not None:
        if (settings_storage.input_file is not None and
                settings_storage.input_file != args.input):
            warnings.warn("Input file in parameters file doesn't match to one"
                          " from -i/--input option. The last is taken.")
        settings_storage.input_file = args.input

    # 3. Checks that we have got all required
    if (settings_storage.input_file is None and
            settings_storage.resume_from is None):
        raise AttributeError("Input file is required. It could be set by "
                             "-i/--input option or via parameters file.")
    if (settings_storage.output_directory is None and
            settings_storage.resume_from is None):
        raise AttributeError("Output directory is required. It could be set "
                             "by -o/--output option or via parameters file.")
    assert settings_storage.output_directory is not None

    ensure_dir_existence(settings_storage.output_directory,
                         check_emptiness=True)

    if settings_storage.custom_filename is not None:
        if (settings_storage.lower_bound is None or
                settings_storage.upper_bound is None):
            raise AttributeError("Please specify either `Lower bound` and "
                                 "`Upper bound` or `Parameter identifiers` "
                                 "for custom model")
    else:
        if (settings_storage.initial_structure is None and
                settings_storage.final_structure is None):
            raise AttributeError("Please specify either structure of "
                                 "demographic history or filename with custom"
                                 " model.")

    if settings_storage.resume_from is not None:
        old_settings = SettingsStorage.from_file(
            os.path.join(settings_storage.resume_from, 'params_file'),
            os.path.join(settings_storage.resume_from, 'extra_params_file'))
        # check what have changed and can we deal with it
        if not settings_storage == old_settings:
            data_settings = ['input_file', 'projections', 'population_labels',
                             'outgroup']
            engine_settings = ['engine', 'pts', 'lower_bound', 'upper_bound',
                               'upper_bound_of_first_split',
                               'upper_bound_of_second_split']
            if_true_settings = ['no_migrations', 'only_sudden',
                                'symmetric_migrations', 'split_fractions']
            forbiden_settings = ['custom_filename', 'initial_structure']

            def differ_in_element(attr_list):
                for attr in attr_list:
                    if getattr(settings_storage, attr) !=\
                            getattr(old_settings, attr):
                        yield attr
            for attr in differ_in_element(data_settings + engine_settings):
                settings_storage.generate_x_transform = True
                if not settings_storage.only_models:
                    warnings.warn(f"Setting {attr} is different in new "
                                  "settings and all likelihoods should be "
                                  "recalculated in new run. Check option only_"
                                  "models maybe it should be set to True.")
                    break
            for attr in differ_in_element(if_true_settings):
                if getattr(settings_storage, attr):
                    settings_storage.generate_x_transform = True
                    if not settings_storage.only_models:
                        warnings.warn(f"Setting {attr} was changed from False "
                                      "to True in new settings and all "
                                      "likelihoods should be recalculated in "
                                      "new run. Check option only_models maybe"
                                      " it should be set to True.")
                    break
            for attr in differ_in_element(forbiden_settings):
                raise ValueError(f"Setting {attr} could not be changed in "
                                 "resumed run.")

    else:
        if settings_storage.only_models:
            warnings.warn("Option `only models`/--only_models  must be used "
                          " --resume option only. It would be ignored.")
    return settings_storage, args



from . import Optimizer, ConstrainedOptimizer, UnconstrainedOptimizer
from .optimizer_result import OptimizerResult
from ..utils import eval_wrapper, ensure_file_existence, fix_args, cache_func

import warnings
import copy
import numpy as np
import scipy
from functools import partial, wraps
import pickle

_registered_local_optimizers = {}


class LocalOptimizer(Optimizer):
    """
    Base class for local optimization.
    See :class:`gadma.optimizers.Optimizer` for more information.
    """
    def optimize(self, f, variables, x0, args=(), options={}, maxiter=None,
                 maxeval=None, verbose=0, callback=None,
                 report_file=None, eval_file=None, save_file=None,
                 restore_file=None, restore_points_only=False,
                 restore_x_transform=None):
        """
        Run optimization of local search algorithm.

        :param f: Target function to optimize.
        :type f: func
        :param variables: Variables of `f` which values should be optimized.
        :type variables: :class:`gadma.utils.VariablePool`
        :param x0: Initial point to start optimization.
        :type x0: list
        :param args: Additional arguments of target function.
        :type args: tuple
        :param options: Additional options kwargs for optimization.
        :type options: dict
        :param maxiter: Maximum number of iterations to run.
        :type maxiter: int
        :param maxeval: Maximum number of evaluations to run. If None then run
                        until converge.
        :type maxeval: int
        :param verbose: Verbosity of the output. If 0 then no reports.
        :type verbose: int
        :param callback: Callback to run after each iteration of optimization.
                         Should be called as `callback(x, y)`
        :type callback: function
        :param report_file: File to save report. Check option `verbose`.
        :type report_file: str
        :param eval_file: File to save all evaluations of the function `f`.
        :type eval_file: str
        :param save_file: File to save information during optimization for its
                          reconstruction.
        :type save_file: str
        :param restore_file: File to restore previous run.
        :type restore_file: str
        :param restore_points_only: Restore point/points from previous run and
                                    run optimization from them once more. If
                                    False then previous run will be resumed.
        :type restore_points_only: bool
        :param restore_x_transform: Restore points but transform them before
                                    usage in this run.
        :type restore_x_transform: function
        """
        raise NotImplementedError


def register_local_optimizer(id, optimizer):
    """
    Registers the specified local optimizer.

    :param id: ID of local optimizer to register.
    :param optimizer: Optimizer to register.
    :type optimizer: :class:`gadma.optimizers.LocalOptimizer`
    """
    if id in _registered_local_optimizers:
        raise ValueError(f"Optimizer '{id}' is already registered.")
    if not isinstance(optimizer, LocalOptimizer):
        raise ValueError("Optimizer is not local.")
    _registered_local_optimizers[id] = optimizer
    optimizer.id = id


def get_local_optimizer(id):
    """
    Returns the local optimizer with the specified id.

    :param id: ID of local optimizer.
    """
    if id not in _registered_local_optimizers:
        raise ValueError(f"Optimizer '{id}' is not registered")
    return copy.deepcopy(_registered_local_optimizers[id])


def all_local_optimizers():
    """
    Returns an iterator over all registered local optimizers.
    """
    for optim in set(_registered_local_optimizers.values()):
        yield copy.deepcopy(optim)


class NoneOptimizer(LocalOptimizer):
    """
    Class that inherits :class:`gadma.optimizers.LocalOptimizer` but do not
    run any optimization.
    """
    def save(self, x0, y, save_file):
        info = (x0, y)
        super(NoneOptimizer, self).save(info, save_file)

    def valid_restore_file(self, save_file):
        try:
            info = self.load(save_file)
        except Exception as e:
            return False
        if not isinstance(info, tuple):
            return False
        if not len(info) == 2:
            return False
        return True

    def optimize(self, f, variables, x0, args=(), options={},
                 linear_constrain=None, maxiter=None, maxeval=None,
                 verbose=0, callback=None, eval_file=None, report_file=None,
                 save_file=None, restore_file=None, restore_points_only=False,
                 restore_x_transform=None):
        prepared_f = self.prepare_f_for_opt(f, args, cache=True)
        wrapped_f = eval_wrapper(prepared_f, eval_file)
        finally_wrapped_f = self.wrap_for_report(wrapped_f, variables,
                                                 verbose, report_file)
        if restore_file is not None and self.valid_restore_file(restore_file):
            x0, y = self.load(restore_file)
            if restore_x_transform is not None:
                x0 = restore_x_transform(x)
                y = None
        if restore_file is None or restore_points_only or y is None:
            y = finally_wrapped_f(x0)
        self.save(x0, y, save_file)
        result = OptimizerResult(x0, y, True, 1, message="SUCCESS",
                                 X=[x0], Y=[y], n_eval=1, n_iter=0)
        return result


register_local_optimizer(None, NoneOptimizer())
register_local_optimizer("None", _registered_local_optimizers[None])


class ScipyOptimizer(LocalOptimizer):
    """
    Class of Scipy local search algorithms.

    :cvar ScipyOptimizer.scipy_methods: List of methods names that are
                                        available.
    :cvar ScipyOptimizer.maxeval_kwarg: List of methods names that support
                                        `maxeval` argument.

    :param method: name of method from :func:`scipy.optimize.minimize`.
    :type method: str
    :param log_transform: If True then values rae optimized in log scale.
    :type log_transform: bool
    :param maximize: If True then maximize target function.
    :type maximize: bool
    """
    scipy_methods = []
    maxeval_kwarg = {}
    opt_type = ''

    def __init__(self, method, log_transform=False, maximize=False):
        if method not in self.scipy_methods:
            raise ValueError(f"There is no such {self.opt_type} method "
                             f"{method} in scipy.minimize. Available methods"
                             f" are: {self.scipy_methods}")
        self.method = method
        super(ScipyOptimizer, self).__init__(log_transform, maximize)

    def save(self, x_best, y_best, n_iter, n_eval, is_finished, save_file):
        """
        Save current achievement in optimization. Dumps `x_best`, `y_best` and
        `is_finished` to `save_file` using :mod:`pickle`.

        :param x_best: Values of best configuration.
        :param y_best: Value of target function on `x_best`.
        :param is_finished: If True then optimization was finished.
        :param save_file: Filename to save data.
        """
        info = (x_best, y_best, int(n_iter), n_eval, is_finished)
        super(ScipyOptimizer, self).save(info, save_file)

    def valid_restore_file(self, save_file):
        try:
            info = self.load(save_file)
        except Exception as e:
            return False
        if not isinstance(info, tuple):
            return False
        if not len(info) == 5:
            return False
        if not isinstance(info[2], int):
            return False
        if not isinstance(info[4], bool):
            return False
        return True

    def prepare_callback(self, f, callback, save_file=None):
        new_callback = super(ScipyOptimizer, self).prepare_callback(callback)

        @wraps(new_callback)
        def wrapper(xk, result_obj=None):
            yk = f(xk)
            r = new_callback(xk, yk)
            if wrapper.x_best is None or wrapper.y_best > yk:
                wrapper.x_best = xk
                wrapper.y_best = yk
            n_eval = None
            if hasattr(f, 'cache_info'):
                n_eval = f.cache_info.misses
            self.save(self.inv_transform(wrapper.x_best),
                      self.sign * wrapper.y_best, wrapper.n_iter, n_eval,
                      False, save_file)
            wrapper.n_iter += 1
            return r

        wrapper.x_best = None
        wrapper.n_iter = 0
        return wrapper

    def get_addit_scipy_kwargs(self, variables):
        raise NotImplementedError

    def optimize(self, f, variables, x0, args=(), options={},
                 linear_constrain=None, maxiter=None, maxeval=None,
                 verbose=0, callback=None, eval_file=None, report_file=None,
                 save_file=None, restore_file=None, restore_points_only=False,
                 restore_x_transform=None):
        """
        Run Scipy optimization.


        :param f: Target function to optimize.
        :type f: func
        :param variables: Variables of `f` which values should be optimized.
        :type variables: :class:`gadma.utils.VariablePool`
        :param x0: Initial point to start optimization.
        :type x0: list
        :param args: Additional arguments of target function.
        :type args: tuple
        :param options: Additional options kwargs for scipy optimization.
        :type options: dict
        :param maxiter: Maximum number of iterations to run.
        :type maxiter: int
        :param maxeval: Maximum number of evaluations to run. If None then run
                        until converge.
        :type maxeval: int
        :param verbose: Verbosity of the output. If 0 then no reports.
        :type verbose: int
        :param callback: Callback to run after each iteration of optimization.
                         Should be called as `callback(x, y)`
        :type callback: function
        :param report_file: File to save report. Check option `verbose`.
        :type report_file: str
        :param eval_file: File to save all evaluations of the function `f`.
        :type eval_file: str
        :param save_file: File to save information during optimization for its
                          reconstruction.
        :type save_file: str
        :param restore_file: File to restore previous run.
        :type restore_file: str
        :param restore_points_only: Restore point/points from previous run and
                                    run optimization from them once more. If
                                    False then previous run will be resumed.
        :type restore_points_only: bool
        :param restore_x_transform: Restore points but transform them before
                                    usage in this run.
        :type restore_x_transform: function
        """
        self.check_variables(variables)
        # Create logging files
        if eval_file is not None:
            ensure_file_existence(eval_file)
        if report_file is not None:
            ensure_file_existence(report_file)
        if save_file is not None:
            ensure_file_existence(save_file)

        is_finished = False
        if restore_file is not None and self.valid_restore_file(restore_file):
            x0, _y, _n_iter, _n_eval, is_finished = self.load(restore_file)
            if restore_x_transform is not None:
                x0 = restore_x_transform(x0)
            if restore_points_only:
                if maxiter is not None:
                    maxiter -= _n_iter
                if maxeval is not None and _n_eval is not None:
                    maxeval -= _n_eval
                is_finished = False

        x0 = np.array(x0, dtype=np.float)

        # Fix args in function f and cache it.
        # TODO: not intuitive solution, think more about it.
        # Fix args (cache we did at the end)
        prepared_f = self.prepare_f_for_opt(f, args, cache=True)
        # Wrap for automatic evaluation log
        wrapped_f = eval_wrapper(prepared_f, eval_file)
        # Wrap for writing report
        finally_wrapped_f = self.wrap_for_report(wrapped_f, variables,
                                                 verbose, report_file)

        f_in_opt = partial(self.evaluate, finally_wrapped_f)
        f_in_opt = fix_args(f_in_opt, (), linear_constrain)

        # Cache our final version of function
#        f_in_opt = cache_func(f_in_opt)
#        print(f_in_opt.cache_info)

        x0_in_opt = self.transform(x0)

        if is_finished or maxiter == 0 or maxeval == 0:
            success = True
            status = 0
            message = "maxiter or maxeval is 0"
            x = x0
            y = self.sign * f_in_opt(x0_in_opt)
            if callback is not None:
                callback(x, y)
            return OptimizerResult(x, y, success, status, message,
                                   [x], [y], 1, 0, X_out=[], Y_out=[])

        self.check_variables(variables)
        options = copy.copy(options)
        if maxiter is not None:
            options['maxiter'] = int(maxiter)
        if maxeval is not None and maxeval > 0:
            if self.method not in self.maxeval_kwarg:
                warnings.warn(f"Local optimization {self.method} do not have"
                              "  an option of max number of evaluations. It "
                              "will be used for maxiter.")
                if maxiter:
                    options['maxiter'] = min(maxeval, maxiter)
                else:
                    options['maxiter'] = maxeval
            else:
                kwarg = self.maxeval_kwarg[self.method]
                options[kwarg] = maxeval
                if kwarg == 'maxiter' and maxiter and maxiter != maxeval:
                    warnings.warn(f"Number of iterations is equal to the "
                                  f"number of function evaluations for "
                                  f"{self.method} optimizer.")
                    options[kwarg] = min(maxeval, maxiter)

        # Create callback for scipy
        if callback is not None:
            if 'callback' in options:
                raise ValueError("You have set two callbacks - first in "
                                 "options and second via callback. Please "
                                 "specify one callback.")
            callback = self.prepare_callback(f_in_opt, callback)

        # IMPORTANT we cannot run scipy optimization for search of 0 params
        if len(variables) == 0:
            x = []
            y = self.sign * f_in_opt(x)
            callback(x, y)
            success = True
            status = 0
            message = "Zero parameters to optimize."
            result = OptimizerResult(x, y, success, status, message,
                                     [x], [y], 1, 1)
            return result

        # Run optimization of SciPy
        addit_kw = self.get_addit_scipy_kwargs(variables)
        res_obj = scipy.optimize.minimize(f_in_opt, x0_in_opt, args=(),
                                          method=self.method, options=options,
                                          callback=callback, **addit_kw)
        # Construct OptimizerResult object to return
        result = OptimizerResult.from_SciPy_OptimizeResult(res_obj)
        result.x = self.inv_transform(result.x)
        result.y = self.sign * result.y

        result.X = [np.array(_x) for _x, _ in prepared_f.cache_info.all_calls]
        result.Y = [self.sign * _y
                    for _, _y in prepared_f.cache_info.all_calls]

        result.n_eval = prepared_f.cache_info.misses
#        print(str(prepared_f.cache_info))

        self.save(result.x, result.y, result.n_iter, result.n_eval,
                  True, save_file)

        return result


class ScipyUnconstrOptimizer(ScipyOptimizer, UnconstrainedOptimizer):
    """
    Base class for Scipy unconstrained optimizations.
    """
    scipy_methods = ['Nelder-Mead', 'Powell', 'CG', 'BFGS', 'Newton-CG',
                     'COBYLA', 'trust-constr', 'dogleg', 'trust-ncg',
                     'trust-exact', 'trust-krylov']
    maxeval_kwarg = {'Nelder-Mead': 'maxfun', 'Powell': 'maxfev',
                     'COBYLA': 'maxiter'}

    opt_type = 'unconstrained'

    def get_addit_scipy_kwargs(self, variables):
        return {}


class ScipyConstrOptimizer(ScipyOptimizer, ConstrainedOptimizer):
    scipy_methods = ['L-BFGS-B', 'TNC', 'SLSQP']
    maxeval_kwarg = {'L-BFGS-B': 'maxfun'}
    opt_type = 'constrained'

    def get_addit_scipy_kwargs(self, variables):
        return {'bounds': [self.transform(var.domain) for var in variables]}


register_local_optimizer('L-BFGS-B', ScipyConstrOptimizer('L-BFGS-B'))
register_local_optimizer('optimize_lbfgsb',
                         _registered_local_optimizers['L-BFGS-B'])
register_local_optimizer('L-BFGS-B_log',
                         ScipyConstrOptimizer('L-BFGS-B',
                                              log_transform=True))
register_local_optimizer('optimize_log_lbfgsb',
                         _registered_local_optimizers['L-BFGS-B_log'])


class ManuallyConstrOptimizer(LocalOptimizer, ConstrainedOptimizer):
    """
    Class for Constrained optimization that uses unconstrained optimization.
    The value of target function is considered to be inf outside bounds.

    :param optimizer: Unconstrained optimizer to use.
    :type optimizer: class:`gadma.optimizers.UnconstrainedOptimizer`
    :param log_transform: If True then log scale is used for optimization.
                          Be careful there could be `log_transform` already
                          in `optimizer`.
    """
    def __init__(self, optimizer, log_transform=False):
        self.optimizer = optimizer
        super(ManuallyConstrOptimizer, self).__init__(log_transform,
                                                      self.optimizer.maximize)
        self.out_of_bounds = np.inf

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, new_id):
        self._id = new_id
        self.optimizer.id = self._id

    @property
    def maximize(self):
        return self.optimizer.maximize

    @maximize.setter
    def maximize(self, new_value):
        self.optimizer.maximize = new_value

    def evaluate_inner(self, f, x, bounds, args=()):
        for val, domain in zip(x, bounds):
            if val < domain[0] or val > domain[1]:
                return self.sign * self.out_of_bounds
        y = f(self.inv_transform(x), *args)
        return y

    def prepare_callback(self, callback):
        @wraps(callback)
        def wrapper(x, y):
            return callback(self.inv_transform(x), y)
        return wrapper

    def valid_restore_file(self, save_file):
        return self.optimizer.valid_restore_file(save_file)

    def optimize(self, f, variables, x0, args=(), options={},
                 linear_constrain=None, maxiter=None, maxeval=None,
                 verbose=0, callback=None, eval_file=None, report_file=None,
                 save_file=None, restore_file=None, restore_points_only=False,
                 restore_x_transform=None):
        self.check_variables(variables)
        x0 = np.array(x0, dtype=np.float)
        x0_in_opt = self.optimizer.transform(self.transform(x0))
        bounds = self.transform([var.domain for var in variables])
        vars_in_opt = copy.deepcopy(variables)
        if isinstance(self.optimizer, UnconstrainedOptimizer):
            for var in vars_in_opt:
                var.domain = np.array([-np.inf, np.inf])
        args_in_opt = (bounds, args)
        # Fix args in function f and cache it.
        # TODO: not intuitive solution, think more about it.
        # Fix args
        prepared_f = self.prepare_f_for_opt(f, args, cache=True)
        # Write automatix evaluation log. Incide optimizer it could different
        # x in function as there can be extra logarithm.
        wrapped_f = eval_wrapper(prepared_f, eval_file)
        # The same with report_file
        finally_wrapped_f = self.wrap_for_report(wrapped_f, variables,
                                                 verbose, report_file)
        f_in_opt = partial(self.evaluate_inner, finally_wrapped_f)
#        f_in_opt = fix_args(f_in_opt, (linear_constrain,))

        if callback is not None:
            callback = self.prepare_callback(callback)
        points_only = restore_points_only
        x_transf = restore_x_transform
        result = self.optimizer.optimize(f_in_opt, vars_in_opt, x0_in_opt,
                                         args=(bounds,), options=options,
                                         verbose=0,
                                         linear_constrain=linear_constrain,
                                         maxiter=maxiter,
                                         maxeval=maxeval, callback=callback,
                                         eval_file=None,
                                         report_file=None,
                                         save_file=save_file,
                                         restore_file=restore_file,
                                         restore_points_only=points_only,
                                         restore_x_transform=x_transf)
        # TODO: need to check result.X as they should be transformed somehow.
        result.x = self.inv_transform(result.x)
        result.X = [self.inv_transform(x) for x in result.X]
        result.n_eval = prepared_f.cache_info.misses
        return result


register_local_optimizer('BFGS',
                         ManuallyConstrOptimizer(
                            ScipyUnconstrOptimizer('BFGS')))
register_local_optimizer('optimize', _registered_local_optimizers['BFGS'])
register_local_optimizer('BFGS_log',
                         ManuallyConstrOptimizer(
                            ScipyUnconstrOptimizer('BFGS'),
                            log_transform=True))
register_local_optimizer('optimize_log',
                         _registered_local_optimizers['BFGS_log'])
register_local_optimizer('Powell',
                         ManuallyConstrOptimizer(
                            ScipyUnconstrOptimizer('Powell')))
register_local_optimizer('optimize_powell',
                         _registered_local_optimizers['Powell'])
register_local_optimizer('Powell_log',
                         ManuallyConstrOptimizer(
                            ScipyUnconstrOptimizer('Powell'),
                            log_transform=True))
register_local_optimizer('optimize_log_powell',
                         _registered_local_optimizers['Powell_log'])
register_local_optimizer('Nelder-Mead',
                         ManuallyConstrOptimizer(
                            ScipyUnconstrOptimizer('Nelder-Mead')))
register_local_optimizer('optimize_fmin',
                         _registered_local_optimizers['Nelder-Mead'])
register_local_optimizer('Nelder-Mead_log',
                         ManuallyConstrOptimizer(
                            ScipyUnconstrOptimizer('Nelder-Mead'),
                            log_transform=True))
register_local_optimizer('optimize_log_fmin',
                         _registered_local_optimizers['Nelder-Mead_log'])



from ..utils import Variable
from .model import Model


class VariablesCombination(Model):
    """
    Base class for combination of variables.
    """
    def __init__(self):
        super(VariablesCombination, self).__init__(raise_excep=False)

    def __str__(self):
        return f"VariablesCombination of {self.variables} variables"


class BinaryOperation(VariablesCombination):
    """
    Combination of two variables.

    :param arg1: First argument.
    :type arg1: :class:`gadma.Variable` or value.
    :param arg2: Second argument.
    :type arg2: :class:`gadma.Variable` or value.

    :raises AssertError: if both arguments are not variables.
    """
    def __init__(self, arg1, arg2):
        super(BinaryOperation, self).__init__()
        self.arg1 = arg1
        self.arg2 = arg2
        self.add_variables([arg1, arg2])
        assert len(self.variables) > 0

    @property
    def name(self):
        """
        Generates name from variables names like:
        `self.arg1.name operation self.arg1.name`
        """
        if isinstance(self.arg1, (Variable, BinaryOperation)):
            arg1_name = self.arg1.name
            if isinstance(self.arg1, BinaryOperation):
                arg1_name = f"({arg1_name})"
        else:
            arg1_name = self.arg1
        if isinstance(self.arg2, (Variable, BinaryOperation)):
            arg2_name = self.arg2.name
            if isinstance(self.arg2, BinaryOperation):
                arg2_name = f"({arg2_name})"
        else:
            arg2_name = self.arg2
        return f"{arg1_name} {self.operation_str()} {arg2_name}"

    def get_value(self, values):
        """
        Returns value of the combination from variables values.

        :param values: Values of the variables.
        :type values: list of dict
        """
        var2value = self.var2value(values)
        if isinstance(self.arg1, BinaryOperation):
            vals = [var2value[var] for var in self.arg1.variables]
            val1 = self.arg1.get_value(vals)
        else:
            val1 = var2value.get(self.arg1, self.arg1)
        if isinstance(self.arg2, BinaryOperation):
            vals = [var2value[var] for var in self.arg2.variables]
            val2 = self.arg2.get_value(vals)
        else:
            val2 = var2value.get(self.arg2, self.arg2)
        return self.operation(val1, val2)

    def string_repr(self, values):
        """
        Returns string representation of combination with defined values.

        :param values: Values of the variables.
        :type values: list of dict
        """
        val = self.get_value(values)
        return f"{self.name}={val}"

    def operation(self, val1, val2):
        """
        Returns the result of binary operation from two values.
        """
        raise NotImplementedError

    def operation_str(self):
        """
        Returns string representation of binary operation.
        """
        raise NotImplementedError


class Addition(BinaryOperation):
    """
    The sum of two variables.
    """
    def operation(self, val1, val2):
        return val1 + val2

    def operation_str(self):
        return "+"


class Subtraction(BinaryOperation):
    """
    The subtraction of two variables.
    """
    def operation(self, val1, val2):
        return val1 - val2

    def operation_str(self):
        return "-"


class Multiplication(BinaryOperation):
    """
    The multiplication of two variables.
    """
    def operation(self, val1, val2):
        return val1 * val2

    def operation_str(self):
        return "*"


class Division(BinaryOperation):
    """
    The division of one variable by another.
    """
    def operation(self, val1, val2):
        return val1 / val2

    def operation_str(self):
        return "/"



import scipy
import numpy as np


def my_dot(A_i, x):
    """
    Implemented easy version of multiplication of two vectors (row and column).
    """
    res = 0
    for _a, _x in zip(A_i, x):
        if _a != 0:
            res += _x * _a
    return res


class LinearConstrain(object):
    """
    Class containing linear constrain. `lb <= A*x <= ub`

    :param A: Matrix.
    :param lb: Lower bound.
    :param ub: Upper bound.
    """
    def __init__(self, A, lb, ub):
        lb = np.nan_to_num(np.array(lb, dtype=np.float), nan=-np.inf)
        ub = np.nan_to_num(np.array(ub, dtype=np.float), nan=np.inf)
        self.constrain = scipy.optimize.LinearConstraint(np.array(A), lb, ub)

    def _get_value(self, x):
        """
        Multiply matrix A by `x`.

        :param x: Vector.
        """
        return np.array([my_dot(A_i, x) for A_i in self.constrain.A])

    def fits(self, x):
        """
        Checks that `x` is good for constrain.

        :param x: Vector to check that `lb <= A*x <=ub`.
        """
        Ax = self._get_value(x)
        lb = self.constrain.lb
        ub = self.constrain.ub
        return (np.all(np.logical_or(Ax > lb, np.isclose(Ax, lb))) and
                np.all(np.logical_or(Ax < ub, np.isclose(Ax, ub))))

    def try_to_transform(self, x):
        """
        Try to transform x to fit the constrain.
        Current implementation go through each pair of bounds and if `x`
        multiplied by the corresponding column of `A` does not fit bounds then
        values of `x` that take part in this constrain are changed so that
        they fit the bounds.

        :param x: Vector to change.

        :returns: transformed x and bool if transformation was successful.
        """
        x_tr = np.array(x)
        for A_i, lb_i, ub_i in zip(self.constrain.A,
                                   self.constrain.lb, self.constrain.ub):
            A_ix = my_dot(A_i, x_tr)
            if A_ix < lb_i:
                C = lb_i
            if A_ix > ub_i:
                C = ub_i
            if A_ix < lb_i or A_ix > ub_i:
                involved = (A_i != 0)
                sign = np.sign(A_i[involved])
                x_tr[involved] = sign * C / np.sum(np.abs(A_i[involved]))
        success = self.fits(x)
        return x_tr, success

    @property
    def lb(self):
        return self.constrain.lb

    @property
    def ub(self):
        return self.constrain.ub

    @property
    def A(self):
        return self.constrain.A

    @lb.setter
    def lb(self, new_value):
        new_value = np.nan_to_num(np.array(new_value, dtype=np.float),
                                  nan=-np.inf)
        self.constrain.lb = new_value

    @ub.setter
    def ub(self, new_value):
        new_value = np.nan_to_num(np.array(new_value, dtype=np.float),
                                  nan=np.inf)
        self.constrain.ub = new_value

    @A.setter
    def A(self, new_value):
        self.constrain.A = new_value

    def __str__(self):
        """
        String representation of constrain.
        """
        lb_size = 0
        ub_size = 0
        A_size = 0
        for lb_i, A_i, ub_i in zip(self.lb, self.A, self.ub):
            lb_size = max(lb_size, len(str(self.lb)))
            ub_size = max(ub_size, len(str(self.ub)))
            A_size = max(A_size, len(str(A_i)))

        lb_str = "{0: <" + str(lb_size) + "}"
        ub_str = "{0: <" + str(ub_size) + "}"
        A_str = "{0: <" + str(A_size) + "}"

        ret_str = "\t".join([lb_str.format("lb"), "|",  A_str.format("A"),
                             "|", ub_str.format("ub")]) + "\n"
        for lb_i, A_i, ub_i in zip(self.lb, self.A, self.ub):
            ret_str += "\t".join([lb_str.format(lb_i), "|",
                                  A_str.format(str(A_i)), "|",
                                  ub_str.format(ub_i)]) + "\n"
        return ret_str


# class LinearConstrainDemographics(LinearConstrain):
#     def __init__(self, A, lb, ub, engine, engine_args):
#         self.engine = engine
#         self.engine_args = engine_args
#         self.original_lb = lb
#         self.original_ub = ub
#         super(LinearConstrainDemographics, self).__init__(A, lb, ub)
#
#     def fits(self, x):
#         theta = self.engine.get_theta(x, *self.engine_args)
#         self.constrain.lb = self.original_lb / theta
#         self.constrain.ub = self.original_ub / theta
#         return super(LinearConstrainDemographics, self).fits(x)
#
#     def try_to_transform(self, x):
#         theta = self.engine.get_theta(x, *self.engine_args)
#         lb = self.original_lb / theta
#         ub = self.original_ub / theta
#         lin_constr = LinearConstrain(self.constrain.A, lb, ub)
#         print("!!!", ub)
#         return lin_constr.try_to_transform(x)



from ..utils import Variable, PopulationSizeVariable, TimeVariable
from ..utils import VariablePool, variables_values_repr
from ..utils import MigrationVariable, DynamicVariable, SelectionVariable
from . import Model, Epoch, Split
from collections import OrderedDict
import copy
import numpy as np


class DemographicModel(Model):
    """
    Base class for demographic model.

    :param gen_time: Time of one generation.
    :type gen_time: float
    :param theta0: Mutation flux (4 * mu * L, where L - length of sequence).
    :type theta0: float
    :param mu: Mutation rate per base per generation.
    :type mu: float
    :param linear_constrain: linear constrain on parameters.
    :type linear_constrain: :class:`gadma.optimizers.LinearConstrain`
    """
    def __init__(self, gen_time=None, theta0=None, mu=None,
                 linear_constrain=None):
        self.gen_time = None
        self.Nref = 1.0
        self.theta0 = theta0  # mutation flux = 4 * mu * length
        self.mu = mu  # mutation rate per base per generation
        self.fixed_vars = {}
        self.linear_constrain = linear_constrain
        super(DemographicModel, self).__init__(raise_excep=False)

    def get_number_of_parameters(self, values):
        """
        Returns number of parameters of the model.
        """
        var2value = self.var2value(values)
        return len(var2value.keys())

    def as_custom_string(self, values):
        """
        Returns string representation of the demographic model with
        parameters as a list with variables and their values.

        :param values: Values of the demographic model.
        """
        if isinstance(values, dict):
            values = [val for var, val in self.var2value(values).items()]
        return variables_values_repr(self.variables, values)

    def translate_units(self, values, Nanc, Tg=1):
        """
        Translates values from genetic units to physical.

        :param values: Values of parameters.
        :param Nanc: Size of ancestral population.
        :param Tg: Time per generation.
        """
        var2value = self.var2value(values)
        translated_values = list()
        for var, val in var2value.items():
            if not hasattr(var, 'translate_units'):
                raise ValueError("Values cannot be translated for demographic"
                                 " model as there is at least one variable "
                                 "that is not specified as demographic.")
            translated_values.append(var.translate_units(val, Nanc))
            if isinstance(var, TimeVariable):
                translated_values[-1] *= Tg
        return translated_values


class EpochDemographicModel(DemographicModel):
    """
    Class for demographic model of epoch type.
    This type is common for :py:mod:`dadi` and :py:mod:`moments`.

    See :class:`gadma.models.DemographicModel` for constructor docs.
    """
    def __init__(self, gen_time=None, theta0=None, mu=None,
                 linear_constrain=None):
        self.events = list()
        super(EpochDemographicModel, self).__init__(gen_time, theta0, mu,
                                                    linear_constrain)

    def _get_current_pop_sizes(self):
        """
        Returns the populations sizes after the last epoch.
        """
        if len(self.events) == 0:
            return [self.Nref]
        return copy.copy(self.events[-1].size_args)

    def number_of_populations(self):
        """
        Returns number of populations in the model.
        """
        return len(self._get_current_pop_sizes())

    def add_epoch(self, time_arg, size_args, mig_args=None,
                  dyn_args=None, sel_args=None, dom_args=None):
        """
        Adds new epoch to the demographic model events.

        :param time_arg: time of the epoch.
        :param size_args: population sizes at the end of the epoch.
        :param mig_args: migrations between populations during the epoch.
        :param dyn_args: dynamics of the populations during the epoch.
        :param sel_args: selection coefficients of the populations during
                         the epoch.
        :param dom_args: dominance coefficients.

        :note: all arguments could contain variables of :class:`Variable`\
               class as well as different constants/values including\
               :class:`gadma.models.BinaryOperation` instances.
        """
        sizes = self._get_current_pop_sizes()
        new_epoch = Epoch(time_arg, sizes, size_args,
                          mig_args, dyn_args, sel_args, dom_args)
        self.events.append(new_epoch)
        self.add_variables(new_epoch.variables)

    def add_split(self, pop_to_div, size_args):
        """
        Adds new split to the demographic model events.

        :param pop_to_div: population to divide.
        :param size_args: population sizes of two subpopulations after the
                          split.
        """
        sizes = self._get_current_pop_sizes()
        sizes[pop_to_div] = size_args[0]
        sizes.append(size_args[1])
        new_split = Split(pop_to_div, sizes)
        self.events.append(new_split)
        self.add_variables(new_split.variables)

    def fix_variable(self, variable, value):
        """
        Fixes variable in the model to the value.

        :param variable: Variable to fix.
        :param value: Value of the variable to fix.
        """
        for event in self.events:
            if variable in event.variables:
                event.fix_variable(variable, value)
                super(DemographicModel, self).fix_variable(variable, value)
                return
        raise ValueError(f"There is no such unfixed variable {variable} in"
                         " demographic model.")

    def unfix_variable(self, variable):
        """
        Unfixes the variable of the model.
        """
        for event in self.events:
            if variable in event._variables:
                event.unfix_variable(variable)
                super(DemographicModel, self).unfix_variable(variable)
                return
        raise ValueError(f"There is no such fixed variable {variable} in"
                         " demographic model.")

    def fix_dynamics(self, values):
        """
        Makes all dynamics in events fixed.

        :param values: Values of parameters to take for fixation.
        :type values: list or dict
        """
        var2value = self.var2value(values)
        assert isinstance(var2value, dict)
        for event in self.events:
            if isinstance(event, Epoch) and event.dyn_args is not None:
                for dyn_arg in event.dyn_args:
                    if isinstance(dyn_arg, Variable):
                        value = var2value[dyn_arg]
                        super(DemographicModel, self).fix_variable(dyn_arg,
                                                                   value)
                        event.fix_variable(dyn_arg, value)

    def unfix_dynamics(self):
        """
        Makes all dynamics in events unfixed.
        """
        for event in self.events:
            if isinstance(event, Epoch) and event.dyn_args is not None:
                for dyn_arg in event.dyn_args:
                    if isinstance(dyn_arg, Variable):
                        super(DemographicModel, self).unfix_if_fixed(dyn_arg)
                        event.unfix_if_fixed(dyn_arg)

    def get_number_of_parameters(self, values):
        """
        Returns number of parameters of the model. Be careful as this number
        could be different to the number of variables. According to the
        population size dynamics in `values` some variables could have no
        influence on the model and in this case they are not counted.

        :param values: Values of the parameters. Dynamics makes the most
                       sense.
        :type values: list or dict
        """
        var2value = self.var2value(values)
        all_variables = VariablePool()
        for event in self.events:
            if isinstance(event, Split):
                continue
            help_event = Epoch(event.time_arg,
                               [None for _ in event.init_size_args],
                               event.size_args,
                               event.mig_args,
                               event.dyn_args,
                               event.sel_args,
                               event.dom_args)
            variables = help_event.variables
            if event.dyn_args is not None:
                for dyn in event.dyn_args:
                    if isinstance(dyn, Variable):
                        value = var2value[dyn]
                    else:
                        value = dyn
                    if value != 'Sud':
                        for init_size in event.init_size_args:
                            if (isinstance(init_size, Variable) and
                                    init_size not in variables):
                                variables.append(init_size)
            for var in variables:
                if var not in all_variables and var not in self.fixed_values:
                    all_variables.append(var)
        return len(all_variables)

    def as_custom_string(self, values):
        """
        Returns string representation of the demographic model with
        parameters.

        :param values: Values of the demographic model.
        """
        strings = []
        values = {var.name: val
                  for var, val in self.var2value(values).items()}
        for event in self.events:
            strings.append(event.as_custom_string(values))
        return "[ " + ",\t".join(strings) + " ]"

    def get_involved_for_split_time_vars(self, n_split):
        """
        Returns list of ints and bias. If value > 0 then this variable is
        involved in sum of times for split.

        It will return A, b: Ax + b = time of `n_split` split.
        """
        var2value = self.var2value(np.zeros(len(self.variables)))
        n_sp = 0
        b = 0
        for event in reversed(self.events):
            if isinstance(event, Split):
                n_sp += 1
                if n_sp == n_split:
                    return list(var2value.values()), b
            else:
                time_arg = event.time_arg
                if isinstance(time_arg, Variable):
                    var2value[time_arg] += 1
                else:
                    b += time_arg
        return [], 0



import operator as op
from functools import partial, wraps
import numpy as np
import copy
import pickle
import sys
import time

import logging
from .optimizer import ConstrainedOptimizer
from .global_optimizer import GlobalOptimizer, register_global_optimizer
from .optimizer_result import OptimizerResult
from ..utils import sort_by_other_list, choose_by_weight, eval_wrapper
from ..utils import ensure_file_existence, fix_args

from .. import GPyOpt
from .. import GPy

logger = logging.getLogger(__name__)


class BayesianOptimizer(GlobalOptimizer, ConstrainedOptimizer):
    """
    Class for Bayesian optimization
    """
    def __init__(self, kernel="Matern52", ARD=True, acquisition_type='MPI',
                 log_transform=False, maximize=False):
        self.kernel_name = kernel
        self.ARD = ARD
        self.acquisition_type = acquisition_type
        super(BayesianOptimizer, self).__init__(log_transform, maximize)
        self.original_transform = copy.deepcopy(self.transform)
        self.original_inv_transform = copy.deepcopy(self.inv_transform)

        def new_transform(x):
            return self.gpyopt_transform(self.original_transform(x))

        def new_inv_transform(x):
            return self.gpyopt_inv_transform(self.original_inv_transform(x))
        self.transform = new_transform
        self.inv_transform = new_inv_transform

    def gpyopt_transform(self, x):
        return [x]

    def gpyopt_inv_transform(self, x):
        return x[0]

    def _get_kernel_class(self):
        return op.attrgetter(self.kernel_name)(GPy.kern.src.stationary)

    def get_kernel(self, ndim):
        kernel = self._get_kernel_class()(ndim, ARD=self.ARD)
        return kernel

    def get_domain(self, variables):
        gpy_domain = []
        for var in variables:
            gpy_domain.append({'name': var.name,
                               'type': var.var_type,
                               'domain': self.original_transform(var.domain)})
        return gpy_domain

#    def evaluate(self, f, x, args=(), linear_constrain=None):
#        return [super(BayesianOptimizer, self).evaluate(f, x[0], args,
#                                                        linear_constrain)]

    def _concatenate_f_and_callback(self, f, callback):
        @wraps(f)
        def concat_wrapper(x):
            y = f(x)
            if callback is not None:
                callback(self.inv_transform(x), y)
            return [y]
        return concat_wrapper

#    def initial_design(self, f, variables, num_init, X_init=None, Y_init=None,
#                       random_type='resample', custom_rand_gen=None):
#        X = list()
#        if X_init is not None:
#            X = list(X_init)
#        for _ in range(num_init - len(X)):
#            x = [self.randomize(variables, random_type, custom_rand_gen)]
#            X.append(x)
#        Y = None
#        if Y_init is not None:
#            Y = np.array(Y).reshape(len(Y), -1)
#        X, Y = super(BayesianOptimizer, self).initial_design(f, variables,
#                                                             num_init, X, Y)
#        return [x[0] for x in X], Y

    def write_report(self, bo_obj, report_file, x, y):
        """
        Writes report about each iteration in file or stdout.

        :param bo_obj: Object of Bayesian Optimization.
        :type bo_obj: GPyOpt.methods.BayesianOptimization
        :param report_file: file to write report to. If None then to stdout.

        :param x: Current solution
        :param y: Current value of fitness function
        """
        bo_obj._compute_results()
        x_best = bo_obj.x_opt
        y_best = bo_obj.fx_opt
        n_iter = bo_obj.num_acquisitions

        if y < y_best:
            x_best = x
            y_best = y

        if report_file is not None:
            stream = open(report_file, 'a')
        else:
            stream = sys.stdout

        print('====================== Iteration %05d ======================' %
              n_iter, file=stream)
        print('Current state of the model:', file=stream)

        print(str(bo_obj.model), file=stream)
        print(bo_obj.model.model.kern.lengthscale, file=stream)
        print('=============================================================',
              file=stream)

        print('*************************************************************',
              file=stream)
        print('Current optimum: %0.3f' % y_best, file=stream)
        print('*************************************************************',
              file=stream)

        if report_file is not None:
            stream.close()

    def optimize(self, f, variables, args=(), num_init=10,
                 X_init=None, Y_init=None,
                 linear_constrain=None, maxiter=100, maxeval=100,
                 verbose=0, callback=None, report_file=None, eval_file=None,
                 save_file=None):
        r"""
        Return best values of `variables` that minimizes/maximizes
        the function `f`.

        :param f: function to minimize/maximize. The usage must be the
                  following: f(x, *args), where x is list of values.
        :param variables: list of variables (instances of
                          :class:`gadma.Variable` class) of the function.
        :param X_init: list of initial values.
        :param Y_init: value of function `f` on initial values from `X_init`.
        :param args: arguments of function `f`.
        :param maxiter: maximum number of genetic algorithm's generations.
        :param maxeval: maximum number of function evaluations.
        :param callback: callback to call after each generation.
                         It will be called as callback(x, y), where x, y -
                         best_solution of generation and its fitness.
        """
        from GPyOpt.methods import BayesianOptimization
        if maxiter is None:
            maxiter = 100
        # Create logging files
        if eval_file is not None:
            ensure_file_existence(eval_file)
        if verbose > 0 and report_file is not None:
            report_file = ensure_file_existence(report_file)
        if save_file is not None:
            ensure_file_existence(save_file)

        # Prepare function to use it.
        # Fix args and cache
        prepared_f = self.prepare_f_for_opt(f, args)
        # Wrap for automatic evaluation logging
        finally_wrapped_f = eval_wrapper(prepared_f, eval_file)

        f_in_opt = partial(self.evaluate, finally_wrapped_f)
        f_in_opt = fix_args(f_in_opt, (), linear_constrain)

        if callback is not None:
            callback = self.prepare_callback(callback)

        # Stuff for BO
        ndim = len(variables)
        if ndim == 0:
            x_best = []
            y_best = f_in_opt([x_best])
            return OptimizerResult(x=x_best, y=self.sign*y_best,
                                   success=True, status="0",
                                   message="Number of variables == 0",
                                   X=[x_best], Y=[y_best],
                                   n_eval=1, n_iter=1,
                                   X_out=[x_best], Y_out=[y_best])

        kernel = self.get_kernel(ndim)
        gpy_domain = self.get_domain(variables)

        # Initial design
        X, Y = self.initial_design(finally_wrapped_f, variables, num_init,
                                   X_init, Y_init)

        Y = np.array(Y).reshape(len(Y), -1)

        bo = BayesianOptimization(f=f_in_opt,
                                  domain=gpy_domain,
                                  model_type='GP',
                                  acquisition_type=self.acquisition_type,
                                  kernel=kernel,
                                  ARD=self.ARD,
                                  X=np.array(X, dtype=object),
                                  Y=np.array(Y),
                                  exact_feval=True,
                                  verbosity=True,
                                  )

        def union_callback(x, y):
            if verbose > 0:
                self.write_report(bo, report_file, x, y)
            if callback is not None:
                callback(x, y)

        f_in_opt = self._concatenate_f_and_callback(f_in_opt, union_callback)

        bo.f = bo._sign(f_in_opt)
        bo.objective = GPyOpt.core.task.objective.SingleObjective(
            bo.f, bo.batch_size, bo.objective_name)

        bo.run_optimization(max_iter=min(maxiter, maxeval)-len(X), eps=0,
                            verbosity=False)

        result = OptimizerResult.from_GPyOpt_OptimizerResult(bo)
        return result


register_global_optimizer('Bayesian_optimization', BayesianOptimizer)
# import torch
# from botorch.models import SingleTaskGP
# from botorch.fit import fit_gpytorch_model
# from botorch.utils import standardize
# from gpytorch.mlls import ExactMarginalLogLikelihood
#
# from botorch.acquisition import ProbabilityOfImprovement
# from botorch.optim import optimize_acqf
#
# import torch
#
# class BOTorchOptimizer(GlobalOptimizer, ConstrainedOptimizer):
#     """
#     Class for default Bayesian optimization with botorch
#     """
#
#     def write_report(self, report_file, x_best, y_best, n_iter):
#         """
#         Writes report about each iteration in file or stdout.
#
#         :param bo_obj: Object of Bayesian Optimization.
#         :type bo_obj: GPyOpt.methods.BayesianOptimization
#         :param report_file: file to write report to. If None then to stdout.
#
#         :param x: Current solution
#         :param y: Current value of fitness function
#         """
#         if report_file is not None:
#             stream = open(report_file, 'a')
#         else:
#             stream = sys.stdout
#
#         print('===================== Iteration %05d =====================' %
#               n_iter, file=stream)
#
#         print('***********************************************************',
#               file=stream)
#         print('Current optimum: %0.3f' % y_best, file=stream)
#         print('***********************************************************',
#               file=stream)
#
#         if report_file is not None:
#             stream.close()
#
#     def optimize(self, f, variables, args=(), num_init=50,
#                  X_init=None, Y_init=None,
#                  linear_constrain=None, maxiter=None, maxeval=None,
#                  verbose=0, callback=None, report_file=None, eval_file=None,
#                  save_file=None, restore_file=None,
#                  restore_points_only=False, restore_x_transform=None):
#         r"""
#         Return best values of `variables` that minimizes/maximizes
#         the function `f`.
#
#         :param f: function to minimize/maximize. The usage must be the
#                   following: f(x, *args), where x is list of values.
#         :param variables: list of variables (instances of
#                           :class:`gadma.Variable` class) of the function.
#         :param X_init: list of initial values.
#         :param Y_init: value of function `f` on initial values from `X_init`.
#         :param args: arguments of function `f`.
#         :param maxiter: maximum number of genetic algorithm's generations.
#         :param maxeval: maximum number of function evaluations.
#         :param callback: callback to call after each generation.
#                          It will be called as callback(x, y), where x, y -
#                          best_solution of generation and its fitness.
#         """
#         if maxiter is None:
#             maxiter = 1000
#         # Create logging files
#         if eval_file is not None:
#             ensure_file_existence(eval_file)
#         if verbose > 0 and report_file is not None:
#             report_file = ensure_file_existence(report_file)
#         if save_file is not None:
#             ensure_file_existence(save_file)
#
#         # Prepare function to use it.
#         # Fix args and cache
#         prepared_f = self.prepare_f_for_opt(f, args)
#         # Wrap for automatic evaluation logging
#         finally_wrapped_f = eval_wrapper(prepared_f, eval_file)
#
#         f_in_opt = partial(self.evaluate, finally_wrapped_f)
#         f_in_opt = fix_args(f_in_opt, (), linear_constrain)
#
#         if callback is not None:
#             callback = self.prepare_callback(callback)
#
#         # Stuff for BO
#         ndim = len(variables)
#         if ndim == 0:
#             x_best = []
#             y_best = f_in_opt(x_best)
#             return OptimizerResult(x=x_best, y=self.sign*y_best,
#                                    success=True, status="0",
#                                    message="Number of variables == 0",
#                                    X=[x_best], Y=[y_best],
#                                    n_eval=1, n_iter=1,
#                                    X_out=[x_best], Y_out=[y_best])
#
#         variables[3].domain = [1e-15, variables[3].domain[1]]
#         bounds = [self.transform([var.domain[0] for var in variables]),
#                   self.transform([var.domain[1] for var in variables])]
#         bounds = torch.stack([torch.FloatTensor(bounds[0]),
#                               torch.FloatTensor(bounds[1])])
#
#         X_init, Y_init = self.initial_design(f_in_opt, variables, num_init,
#                                              X_init, Y_init)
#         Y_total = Y_init
#         X_total = [self.transform(x) for x in X_init]
#
#         y_best = min(Y_total)
#         x_best = X_total[Y_total.index(y_best)]
#
#         for n_iter in range(maxiter):
#             if (maxeval is not None and
#                     finally_wrapped_f.cache_info.missed < maxeval):
#                 break
#             train_X = torch.from_numpy(np.array(X_total, dtype=float))
#             train_Y = standardize(torch.from_numpy(
#                 np.array([[self.sign * y] for y in Y_total], dtype=float)))
#             gp = SingleTaskGP(train_X, train_Y)
#             mll = ExactMarginalLogLikelihood(gp.likelihood, gp)
#             mll = fit_gpytorch_model(mll)
#
#             MPI = ProbabilityOfImprovement(mll.model, best_f=y_best,
#                                            maximize=True)
#             n_repeats = 5
#             initial_points = [[self.transform(
#                 [var.resample() for var in variables])]
#                  for _ in range(n_repeats)]
#             import dadi
#             for _ in range(n_repeats):
#                 x = self.inv_transform(x_best)
#                 lb = [var.domain[0] for var in variables]
#                 ub = [var.domain[1] for var in variables]
#                 new_x = dadi.Misc.perturb_params(x, fold=1e-15,
#                                                  lower_bound=lb,
#                                                  upper_bound=ub)
#                 initial_points.append([self.transform(new_x)])
#             print([f_in_opt(x[0]) for x in initial_points])
#             initial_points = torch.from_numpy(np.array(initial_points,
#                                                        dtype=float))
#             candidate, acq_value = optimize_acqf(
#                 MPI, bounds=bounds, q=1, num_restarts=2*n_repeats,
#                 raw_samples=20, batch_initial_conditions=initial_points)#,
# #                                                 return_best_only=False)
#             print(acq_value)
#
#             X_total.append(candidate[0].tolist())
#             Y_total.append(f_in_opt(X_total[-1]))
#             if Y_total[-1] < y_best:
#                 x_best = X_total[-1]
#                 y_best = Y_total[-1]
#             print(n_iter, X_total[-1], Y_total[-1], y_best)
#             self.write_report(report_file, self.inv_transform(x_best),
#                               self.sign * y_best, n_iter)
#
#         if n_iter == maxiter:
#             message = "Maximum number of iterations achieved"
#         else:
#             message = "Maximum number of evaluations achieved"
#         x_best = self.inv_transform(x_best)
#         y_best = self.sign * y_best
#         X_total = [self.inv_transform(x) for x in X_total]
#         Y_total = [self.sign * y for y in Y_total]i
#         n_eval = finally_wrapped_f.cache_info.misses
#         result = OptimizerResult(x=x_best, y=y_best, success=True,
#                                  status="1", message=message, X=X_total,
#                                  Y=Y_total, n_eval=n_eval, n_iter=n_iter)
#
#         return result
#
# register_global_optimizer('BOTorch', BOTorchOptimizer)
#



from .model import Model
from .event import Event, Epoch, Split
from .demographic_model import DemographicModel, EpochDemographicModel
from .structure_demographic_model import StructureDemographicModel
from .custom_demographic_model import CustomDemographicModel
from .variables_combinations import VariablesCombination
from .variables_combinations import BinaryOperation, Addition, Subtraction,\
                                    Multiplication, Division



try:
    import matplotlib
    matplotlib.use("Agg")
except ImportError:
    pass

try:
    import PIL
    from PIL import Image
except ImportError:
    PIL = None
    Image = None
try:
    import matplotlib
except ImportError:
    matplotlib = None
try:
    import moments
except ImportError:
    moments = None
try:
    import dadi
except ImportError:
    dadi = None
try:
    import GPy
except ImportError:
    GPy = None
try:
    import GPyOpt
except ImportError:
    GPyOpt = None
import warnings

PIL_available = PIL is not None
matplotlib_available = matplotlib is not None
moments_available = moments is not None
dadi_available = dadi is not None
GPy_available = GPy is not None
GPyOpt_available = GPyOpt is not None

from .data import *  # NOQA
from .engines import *  # NOQA
from .models import *  # NOQA
from .code_generator import id2printfunc  # NOQA
from .utils import *  # NOQA
from .optimizers import *  # NOQA
from .cli import *  # NOQA
from .core import *  # NOQA
from .Inference import *  # NOQA
from .run_ls_on_boot_data import *  # NOQA
from .get_confidence_intervals import *  # NOQA

warnings.simplefilter('always', UserWarning)
warnings.filterwarnings(action='ignore', category=UserWarning,
                        module='matplotlib')
warnings.filterwarnings(action='ignore', category=UserWarning,
                        module='moments')
warnings.filterwarnings(action='ignore', category=UserWarning,
                        module='dadi')
warnings.formatwarning = utils.warning_format



from .variables import Variable, ContinuousVariable, DiscreteVariable, \
    TimeVariable, PopulationSizeVariable, MigrationVariable, \
    SelectionVariable, DynamicVariable, FractionVariable
from .variable_pool import VariablePool
from .utils import *
from .distributions import *



from functools import wraps
import time
import numpy as np
from collections import namedtuple
import sys
import os
import copy


def logarithm_transform(x):
    """
    Transforms ``x`` by applying ``numpy.log`` on it.
    """
    if isinstance(x, (int, float, np.integer, np.float)):
        x = [x]
    if not isinstance(x, (list, np.ndarray)):
        return x
    if isinstance(x, np.ndarray):
        x = x.astype(float)
    return np.log(x)


def exponent_transform(x):
    """
    Transforms ``x`` by applying ``numpy.exp`` on it.
    """
    if isinstance(x, (int, float, np.integer, np.float)):
        x = [x]
    if not isinstance(x, (list, np.ndarray)):
        return x
    if isinstance(x, np.ndarray):
        x = x.astype(float)
    return np.exp(x)


def ident_transform(x):
    """
    Identical transform. Just returns ``x``.
    """
    return x


def choose_by_weight(X, weights, nsample):
    """
    Choose ``nsample`` samples from ``X`` according to ``weights``.
    The greater weight is the greater the probability to choose sample is.

    :note: if weights is None then choice will be uniform
    """
    if weights is None:
        weights = np.ones(len(X))
    p = np.array(weights)
    p /= np.sum(p)
    return np.random.choice(X, size=nsample, replace=False)


def sort_by_other_list(x, y, reverse=False, key=None):
    """
    Sort ``x`` and ``y`` according to values in ``y``.
    """
    def ident(x):
        return x
    if key is None:
        key = ident
    sort_zip = sorted(zip(x, y), key=lambda p: key(p[1]))
    return [p[0] for p in sort_zip], [p[1] for p in sort_zip]


def fix_args(f, *args):
    r"""
    Fixes argumets of function.

    :param f: Function such that f(x, *args)
    :param args: Tuple of function arguments.

    :returns: function that will take only x as argument.
    """
    @wraps(f)
    def fix_args_wrapper(x):
        return f(x, *args)
    return fix_args_wrapper


class CacheInfo(object):
    """
    Class for keeping cache info like one from ``functools.lru_cache``
    cache_info.
    """
    def __init__(self):
        self.hits = 0
        self.misses = 0
        self.cache = {}
        self.all_calls = []

    def __str__(self):
        return f"hits: {self.hits}, misses: {self.misses}"


def lru_cache(func):
    """
    Our lru cache. We want to get cache itself while functools.lru_cache
    could not do it.
    Please, be carefull as it could be some attributes named the same way
    and it will be ruined. We use it for our decorator :func:`cache_func`.
    """
    func.cache_info = CacheInfo()

    @wraps(func)
    def lru_cache_wrapper(*args):
        try:
            ret = func.cache_info.cache[args]
            func.cache_info.hits += 1
        except KeyError:
            ret = func(*args)
            func.cache_info.cache[args] = ret
            func.cache_info.misses += 1
        func.cache_info.all_calls.append([args, ret])
        return ret
    return lru_cache_wrapper


def cache_func(f):
    """
    Cashes function with one argument.
    :param f: function such that f(x).
    :returns: function that is cashed.
    """
    @lru_cache
    @wraps(f)
    def tuple_wrapper(x_tuple):
        return f(np.array(x_tuple, dtype=object))

    @wraps(tuple_wrapper)
    def cache_wrapper(x):
        return tuple_wrapper(tuple(x))

    cache_wrapper.cache_info = tuple_wrapper.cache_info
    return cache_wrapper


# def cache_func_2d(f):
#     """
#     Cashes function with two arguments.
#     :param f: function such that f(x, y).
#     :returns: function that is cashed.
#     """
#     @lru_cache
#     @wraps(f)
#     def tuple_wrapper(x_y_tuple):
#         x, y = x_y_tuple
#         if isinstance(y, tuple) and isinstance(y[0], tuple):
#             y = np.array(y)
#         return f(x, y)
#
#     @wraps(tuple_wrapper)
#     def cache_wrapper(x, y):
#         y_tuple = tuple(y)
#         if (isinstance(y, (list, np.ndarray)) and
#                 isinstance(y[0], (list, np.ndarray))):
#             for i in range(len(y_tuple)):
#                 y_tuple[i] = tuple(y_tuple[i])
#         return tuple_wrapper(tuple(x), y_tuple)
#
#     cache_wrapper.cache_info = tuple_wrapper.cache_info
#     return cache_wrapper


def nan_fval_to_inf(f):
    """
    Wrappes function to return infinity instead nan.
    """
    @wraps(f)
    def nan_fval_to_inf_wrapper(x):
        y = f(x)
        if y is None or np.isnan(y):
            return np.inf
        return y
    return nan_fval_to_inf_wrapper


def eval_wrapper(f, eval_file=None):
    r"""
    Returns good function for optimization. Each evaluation of function will
    be written in file. If needed function will be cached.

    :param f: function. Is called as f(x, *args).
    :param args: tuple of arguments.
    :param eval_file: file to write evaluations.
    :param cache: if True then function will be cached.
    """
    time_init = time.time()
    first_line = '\t'.join(['Time of evaluation start', 'Function value',
                            'Parameters values', 'Evaluation time'])
    if eval_file is not None:
        if not check_file_existence(eval_file):
            open(eval_file, 'w').close()
        if not os.stat(eval_file).st_size == 0:
            with open(eval_file, 'r') as fl:
                line = next(fl)
        if os.stat(eval_file).st_size == 0 or line.strip() != first_line:
            with open(eval_file, 'a') as fl:
                print(first_line, file=fl, sep='\t')

    @wraps(f)
    def eval_wrapper_f(x):
        time_start = time.time()
        y = f(x)
        time_end = time.time()
        if eval_file is not None:
            with open(eval_file, 'a') as fl:
                print(time_start - time_init, y, list(x),
                      time_end - time_start, file=fl, sep='\t')
        return y
    return eval_wrapper_f


class WeightedMetaArray(np.ndarray):
    """
    Array with metadata.

    :param array: array to keep.
    :param dtype: dtype of elements of the array.
    :param order: see ``numpy.ndarray`` for more information.
    """
    def __new__(cls, array, dtype=None, order=None):
        obj = np.asarray(np.array(array, dtype=object),
                         dtype=dtype, order=order).view(cls)
        obj.metadata = ""
        obj.weights = np.ones(obj.shape)
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.metadata = getattr(obj, 'metadata', [{}]*(obj.ndim+1))
        self.weights = getattr(obj, 'weights', [{}]*(obj.ndim+1))

    def __str__(self):
        super_str = super(WeightedMetaArray, self).__str__()
        if hasattr(self, 'metadata'):
            return super_str + '\t' + self.metadata
        return super_str

    def __repr__(self):
        super_str = super(WeightedMetaArray, self).__repr__()
        if hasattr(self, 'metadata'):
            return super_str + '\t' + self.metadata
        return super_str

#    def serialize(self):
#        return (self, self.weights, self.metadata)
#
#    def deserialize(self, data):
#        self = WeightedMetaArray(data[0])
#        self.weights = data[1]
#        self.metadata = data[2]


def list_with_weights_for_pickle(X):
    """
    Transforms elements of ``X`` to pickle it.
    """
    new_X = list()
    for x in X:
        if isinstance(x, WeightedMetaArray):
            new_X.append((x, x.weights, x.metadata))
        else:
            new_X.append(x)
    return new_X


def list_with_weights_after_pickle(X):
    """
    Transforms back pickles version of ``X``.
    """
    new_X = list()
    for x in X:
        if isinstance(x, tuple):
            arr = WeightedMetaArray(x[0])
            arr.weights = x[1]
            arr.metadata = x[2]
            new_X.append(arr)
        else:
            new_X.append(x)
    return new_X


def update_by_one_fifth_rule(value, const, was_improved):
    """
    Updates ``value`` according to 'one-fifth' rule and ``const``. Used in
    genetic algorithm.

    :param value: Value to change.
    :param const: Const for rule.
    :param was_improved: Bool if fitness was improved or not.
    """
    if was_improved:
        return value * const
    return value / (const) ** (0.25)


def abspath(path):
    return os.path.abspath(os.path.expanduser(path))


def check_file_existence(path_to_file):
    return os.path.exists(path_to_file) and os.path.isfile(path_to_file)


def check_dir_existence(path_to_dir):
    return os.path.exists(path_to_dir) and os.path.isdir(path_to_dir)


def ensure_file_existence(path_to_file):
    filename = abspath(path_to_file)
    if not check_file_existence(filename):
        open(filename, 'w').close()
    return filename


def ensure_dir_existence(path_to_dir, check_emptiness=False):
    dirname = abspath(path_to_dir)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    if os.listdir(dirname) != [] and check_emptiness:
        raise RuntimeError(f"Directory {path_to_dir} is not empty\nYou can "
                           f"write:  rm -rf {dirname}\t to remove directory.")
    return dirname


class StdAndFileLogger(object):
    """
    Logger for printing output both in file and stdout.
    """
    def __init__(self, log_filename, silent=False):
        self.terminal = sys.stdout
        self.log_filename = log_filename
        self.silent = silent
        if not os.path.exists(self.log_filename):
            open(self.log_filename, 'w'). close()

    def write(self, message):
        if not self.silent:
            self.terminal.write(message)
            self.terminal.flush()
        with open(self.log_filename, 'a') as fl:
            fl.write(message)

    def flush(self):
        # this flush method is needed for python 3 compatibility.
        # this handles the flush command by doing nothing.
        # you might want to specify some extra behavior here.
        pass


def get_aic_score(n_params, log_likelihood):
    """
    Returns AIC score.

    :param n_params: Number of parameters of model.
    :param log_likelihood: Value of log likelihood.
    """
    return 2 * n_params - 2 * log_likelihood


def get_claic_score(engine, x0, boots,
                    args=(), log_likelihood=None, return_eps=False):
    """
    Calculate CLAIC score for the model.

    :param engine: Engine with model and data.
    :param x0: Parameters of the model.
    :param boots: Bootstrap data.
    :params args: Arguments for engine's ``evaluate`` function.
    :params log_likelihood: Value of log-likelihood for ``x0``.
    :params return_eps: If True then tuple (CLAIC, eps) is returned.
    """
    if log_likelihood is None:
        log_likelihood = engine.evaluate(x0, *args)
    eps = 1e-5
    claic_score = None
    while eps <= 1e-2:
        try:
            claic_component = engine.get_claic_component(x0, boots,
                                                         *args, eps)
            claic_score = 2 * claic_component - 2 * log_likelihood
            break
        except np.linalg.linalg.LinAlgError as e:
            if str(e) == 'Singular matrix':
                eps *= 10
            else:
                print(e)
                raise e
    if return_eps:
        return claic_score, eps
    else:
        return claic_score


# Printing functions
def float_repr(value, precision=5):
    if value < 10**(-precision):
        return f"{value:.2e}"
    return f"{round(value, precision)}"


def variables_values_repr(variables, values):
    val_repr = [float_repr(val) if isinstance(val, float) else val
                for val in values]
    var_val = zip(variables, val_repr)
    x_repr = ",\t".join([f"{var.name}={val}" for var, val in var_val])
    x_repr = f"({x_repr})"
    return x_repr


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def warning_format(message, category, filename, lineno, file=None, line=None):
    return f"{bcolors.WARNING}{category.__name__}: {message}"\
           f"{bcolors.ENDC} ({filename}:{lineno})\n"



import scipy
import numpy as np
import copy
from .. import GPyOpt


class OptimizerResult(object):
    """
    Class for keeping optimizers result.
    It is based on SciPy.optimize.OptimizeResult but have more information.

    :param x: The solution of the optimization. The best value during run.
    :param y: The value of objective function on x.
    :param success: Whether or not the optimizer exited successfully.
    :type success: bool
    :param status: Termination status of the optimizer. Its value depends on
                   the underlying solver. Refer to message for details.
    :type status: int
    :param message: Description of the cause of the termination.
    :type message: str
    :param X: All solutions that were used in run.
    :param Y: Values of objective function on X.
    :param n_eval: Number of evaluations of the objective functions performed
                   by the optimizer.
    :type n_eval: int
    :param n_iter: Number of iterations performed by the optimizer.
    :type n_iter: int
    :param X_out: Solutions that optimizer decides to show as important at
                  the end.
                  For example, it could be solutions of the last iteration.
                  The most usefull usage is to restart optimizer or run new
                  optimizer with passing it via X_init.
    :param Y_out: Values of objective function on X_out.

    """
    def __init__(self, x, y, success: bool, status: int, message: str,
                 X, Y, n_eval: int, n_iter: int, X_out=[], Y_out=[]):
        self.x = np.array(x, dtype=object)
        self.y = y
        self.success = success
        self.status = status
        self.message = message
        self.X = copy.deepcopy(X)
        self.Y = Y
        self.n_eval = n_eval
        self.n_iter = n_iter
        self.X_out = copy.deepcopy(X_out)
        self.Y_out = Y_out

    @staticmethod
    def from_SciPy_OptimizeResult(
            scipy_result: scipy.optimize.OptimizeResult):
        """
        Create OptimizerResult from instance of SciPy.optimize.OptimizeResult.
        Please, note that some attributes will be empty.
        """
        return OptimizerResult(x=scipy_result.x,
                               y=scipy_result.fun,
                               success=scipy_result.success,
                               status=scipy_result.status,
                               message=scipy_result.message,
                               X=[],
                               Y=[],
                               n_eval=scipy_result.nfev,
                               n_iter=scipy_result.nit,
                               X_out=[scipy_result.x],
                               Y_out=[scipy_result.fun])

    @staticmethod
    def from_GPyOpt_OptimizerResult(gpyopt_obj):
        """
        Create OptimizerResult from instance of bayesian optimization.

        :param gpyopt_obj: Object of GPyOpt optimizer
        :type gpyopt_obj: GPyOpt.methods.BayesianOptimization
        """
        gpyopt_obj._compute_results()
        if (gpyopt_obj.num_acquisitions == gpyopt_obj.max_iter and
                not gpyopt_obj.initial_iter):
            message = '   ** Maximum number of iterations reached **'
            success = True
            status = 1
        elif (gpyopt_obj._distance_last_evaluations() < gpyopt_obj.eps and
                not gpyopt_obj.initial_iter):
            message = '   ** Two equal location selected **'
            success = True
            status = 1
        elif (gpyopt_obj.max_time < gpyopt_obj.cum_time and
                not gpyopt_obj.initial_iter):
            message = '   ** Evaluation time reached **'
            success = True
            status = 0
        else:
            message = '** GPyOpt Bayesian Optimization class initialized '\
                      'successfully **'
            success = False
            status = 2

        if hasattr(gpyopt_obj.f, 'cache_info'):
            n_eval = gpyopt_obj.f.cache_info.misses
        else:
            n_eval = len(gpyopt_obj.Y)

        return OptimizerResult(x=gpyopt_obj.x_opt,
                               y=gpyopt_obj.fx_opt,
                               success=success,
                               status=status,
                               message=message,
                               X=gpyopt_obj.X,
                               Y=gpyopt_obj.Y,
                               n_eval=n_eval,
                               n_iter=gpyopt_obj.num_acquisitions,
                               X_out=gpyopt_obj.X,
                               Y_out=gpyopt_obj.Y)

    def __repr__(self):
        string = f"  status: {self.status}\n"\
                 f" success: {self.success}\n"\
                 f" message: {self.message}\n"\
                 f"       x: {self.x}\n"\
                 f"       y: {self.y}\n"\
                 f"  n_eval: {self.n_eval}\n"\
                 f"  n_iter: {self.n_iter}\n"
        return string



from .engine import Engine, get_engine, all_engines, register_engine
from .dadi_engine import DadiEngine
from .moments_engine import MomentsEngine
from .dadi_moments_common import DadiOrMomentsEngine



from ..utils import WeightedMetaArray
from ..engines import get_engine
import copy
from multiprocessing import Manager
from functools import partial
from collections import OrderedDict
import numpy as np


class SharedDict(object):
    """
    Wrapper class on ``multiprocessing.Manager.dict`` that could be used for
    multiprocessing applications. Is used as shared memory of processes that
    are run in parallel for GADMA.

    All models in this dict can be divided in several processes and in several
    groups. When new model is added it has its own process and group. When
    dict should return some models it sort them by some value that is defined
    by `key` function and could be specific for each group.

    :param multiprocessing: If False than usual dict will be used inside.
    """
    def __init__(self, multiprocessing=True):
        if multiprocessing:
            self.dict = Manager().dict()
        else:
            self.dict = dict()

    def default_key(self, group):
        """
        Returns function for `key` in sort. Sorted elements will be compared by
        the value of `default_key(element)`.

        :param group: Name of group of models.
        """
        return None

    @staticmethod
    def get_value(model, key):
        """
        Returns value of `key` from model.

        :param model: Model.
        :param key: Key function.
        """
        if key is None:
            return model
        return key(model)

    def _put_new_model_for_process(self, process, group, model, key=None):
        if key is None:
            key = self.default_key(group)
        new_value = self.get_value(model, key)

        copy_dict = dict(self.dict)
        try:
            process_dict = OrderedDict(copy_dict[process])
        except KeyError:
            process_dict = OrderedDict()
        process_dict[group] = copy.deepcopy(model)
        self.dict[process] = process_dict

    def update_best_model_for_process(self, process, group, model, key=None):
        """
        Updates best model for process. Models are compared by the value of
        `key` function. So if new model has greater value of `key` function
        then it is saved in the dictionary as best for this process in this
        group.

        :param process: Name of process.
        :param group: Name of model group.
        :param model: Model.
        :param key: `key` function, if None then :meth:`default_key` is used.
        """
        if key is None:
            key = self.default_key(group)
        new_value = self.get_value(model, key)

        copy_dict = dict(self.dict)
        try:
            process_dict = OrderedDict(copy_dict[process])
        except KeyError:
            process_dict = OrderedDict()

        if group in process_dict:
            old_model = self.get_best_model_for_process_in_group(process,
                                                                 group)
            old_value = self.get_value(old_model, key)
        if (group not in process_dict or old_value is None or
                np.allclose(new_value, old_value) or new_value >= old_value):
            self._put_new_model_for_process(process, group, model, key)
            return True
        return False

    def add_model_for_process(self, process, group, model, key=None):
        """
        Adds additional model for process. After this method dict will have
        list of models saved under (process, group).

        :param process: Name of process.
        :param group: Name of model group.
        :param model: Model.
        :param key: `key` function, if None then :meth:`default_key` is used.
        """
        if key is None:
            key = self.default_key(group)
        copy_dict = dict(self.dict)
        try:
            process_dict = OrderedDict(copy_dict[process])
        except KeyError:
            process_dict = OrderedDict()
        if group in process_dict:
            models = self.get_models_for_process_in_group(process, group)
        else:
            models = []
        models.append(model)
        process_dict[group] = copy.deepcopy(models)
        self.dict[process] = process_dict

    def get_models_for_process_in_group(self, process, group, key=None):
        """
        Returns models for (process, group).
        """
        if key is None:
            key = self.default_key(group)
        try:
            process_dict = OrderedDict(self.dict[process])
        except KeyError:
            process_dict = OrderedDict()
        if group not in process_dict:
            return []
        if isinstance(process_dict[group], (list, np.ndarray)):
            return sorted(process_dict[group], key=key, reverse=True)
        return [process_dict[group]]

    def get_models_in_group(self, group, key=None):
        """
        Returns dict of sorted pairs (process, model) for group
        (across all processes).
        """
        if key is None:
            key = self.default_key(group)
        models = []
        for process in self.dict.keys():
            process_models = self.get_models_for_process_in_group(process,
                                                                  group,
                                                                  key=key)
            if len(process_models) > 0:
                for model in process_models:
                    models.append([process, model])
            models = sorted(models, key=lambda x: key(x[1]), reverse=True)
        return models

    def get_best_model_in_group(self, group, key=None):
        """
        Returns best model for group.
        """
        if key is None:
            key = self.default_key(group)
        models = self.get_models_in_group(group, key)
        if len(models) == 0:
            return None
        return models[0][1]

    def get_best_model_for_process_in_group(self, process, group, key=None):
        """
        Returns best model for process, group.
        """
        if key is None:
            key = self.default_key(group)
        models = self.get_models_for_process_in_group(process, group, key)
        if len(models) == 0:
            return None
        return models[0]

    def get_available_groups(self):
        """
        Returns all available groups across processes.
        """
        names = []
        copy_dict = dict(self.dict)
        for process in copy_dict.keys():
            for group in copy_dict[process]:
                if group not in names:
                    names.append(group)
        return names


class SharedDictForCoreRun(SharedDict):
    """
    Class for shared dict in :class:`gadma.core.core_run.CoreRun`.

    Process is name of the process or index of CoreRun. Group is name of
    fitness function: log-likelihood, AIC, CLAIC. Model is tuple of
    demographic model, engine and fitness for this engine.
    """
    def default_key(self, group):
        """
        For not `log-likelihood` groups sort should be reversed so `key`
        function multiplies fitness by -1. Also fitness could be dict of
        several values of several groups (group of `log-likelihood` but
        fitness has also value of AIC or CLAIC). So this function extracts
        correct fitness.

        :param group: Name of fitness function (log-likelihood, AIC, CLAIC).
        """
        return partial(self._key, group)

    def construct_model(self, group, engine, x, y):
        """
        Constructs model for group, engine and x, y. Model is a tuple of
        (engine, x, fitness), where engine contain demographic model and
        fitness is dict with available fitnesses (logLL, AIC, CLAIC).

        :param group: Name of fitness (log-likelihood, AIC, CLAIC)
        :param engine: Engine with dem. model and data.
        :param x: Values of dem. model parameters.
        :param y: Value of fitness defined by `group`.
        """
        if not isinstance(y, dict):
            y = OrderedDict({group: y})
        # print(type(x), x)
        if isinstance(x, WeightedMetaArray):
            return (engine, (x, x.metadata), y)
        return (engine, x, y)

    def _extract_model(self, model):
        """
        Extract engine, x, y from the model in shared dict. Reverse function
        to :meth:`construct_model`.
        """
        engine, x_with_metadata, y = model
        if isinstance(x_with_metadata, tuple):
            x = WeightedMetaArray(x_with_metadata[0])
            x.metadata = x_with_metadata[1]
        else:
            x = x_with_metadata
        return engine, x, y

    def _key(self, group, model):
        """
        See :meth:`default_key`
        """
        sign = -1
        if group == 'log-likelihood':
            sign = 1
        _1, _2, y = self.construct_model(group, *model)
        if isinstance(y[group], tuple):
            ff = y[group][0]
        else:
            ff = y[group]
        if ff is None:
            return ff
        return sign * ff

    def _put_new_model_for_process(self, process, group, model, key=None):
        if isinstance(model, tuple):
            engine, x, y = model
        model = self.construct_model(group, engine, x, y)
        r = super(SharedDictForCoreRun, self)._put_new_model_for_process(
            process, group, model)
        return r

    def update_best_model_for_process(self, process, group, engine, x, y):
        # print(x)
        model = self.construct_model(group, engine, x, y)
        r = super(SharedDictForCoreRun, self).update_best_model_for_process(
            process, group, model)
        return r

    def add_model_for_process(self, process, group, engine, x, y):
        model = self.construct_model(group, engine, x, y)
        r = super(SharedDictForCoreRun, self).add_model_for_process(
            process, group, model)
        return r

    def get_models_in_group(self, group, key=None, align_y_dict=False):
        """
        Returns models for specified group.

        :param group: Name of fitness (log-likelihood, AIC, CLAIC).
        :param key: `Key` function.
        :param align_y_dict: If True then all fitnesses are dict with keys
                             across all available groups and None's if value
                             for some group is not available.
        """
        models = super(SharedDictForCoreRun, self).get_models_in_group(group,
                                                                       key)
        if not align_y_dict:
            return models
        y_dict_keys = list()
        for process, model in models:
            for k in model[2]:
                if k not in y_dict_keys:
                    y_dict_keys.append(k)
        for i, (process, model) in enumerate(models):
            for k in y_dict_keys:
                if k not in model[2]:
                    models[i][1][2][k] = None
                models[i][1] = self._extract_model(models[i][1])
        return models



from .optimizer import *
from .local_optimizer import *
from .global_optimizer import *
from .genetic_algorithm import *
from .bayesian_optimization import *
from .combinations import *
from .linear_constrain import LinearConstrain



from ..engines import get_engine, all_engines
from ..utils import sort_by_other_list, ensure_dir_existence,\
                    ensure_file_existence
from ..utils import TimeVariable, PopulationSizeVariable, SelectionVariable,\
                    DynamicVariable
from ..optimizers import GlobalOptimizerAndLocalOptimizer
from ..utils import get_aic_score, get_claic_score, ident_transform, bcolors
from ..models import EpochDemographicModel, StructureDemographicModel
from .draw_and_generate_code import draw_plots_to_file, generate_code_to_file
from ..cli import SettingsStorage
import os
import numpy as np
from collections import defaultdict, OrderedDict
import copy
import warnings
from functools import partial


class CoreRun(object):
    """
    Class of main run in GADMA.
    Has a :meth:`run` method to start launch.
    Runs creates new directory named by its `index` in the output directory
    then all log, code and pictures will be saved there. AIC and CLAIC is
    calculated here.

    :param index: Index of the run. Like id.
    :type index: int
    :param shared_dict: Dictionary to save results in callbacks. Will be saved
                        with key equal to `index`. Is used for
                        multiprocessing cooperation.
    :type shared_dict: :class:`gadma.cor.shared_dict.SharedDictForCoreRun`
    :param settings: Settings of the run. Information to form output directory
                     and so on will be taken from settings.
    :type settings: :class:`gadma.cli.settings_storage.SettingsStorage`
    """
    REPORT_FILENAME = 'GADMA_GA.log'
    EVAL_FILENAME = 'eval_file'
    SAVE_FILENAME = 'save_file'

    def __init__(self, index, shared_dict, settings):
        # 1. Save all init arguments
        self.index = index
        self.shared_dict = shared_dict
        self.settings = settings

        # 2. Extract other information from settings
        # 2.1 Take engine, data and model for the first start.
        self.engine = get_engine(self.settings.engine)
        self.data = self.settings.inner_data
        self.model = self.settings.get_model()
        # We save data_holder in engine for good code generation
        self.engine.data_holder = self.settings.data_holder
        # 2.2 Get optimizers and their kwargs that will be used.
        self.global_optimizer = self.settings.get_global_optimizer()
        self.local_optimizer = self.settings.get_local_optimizer()
        self.optimize_kwargs = self.settings.get_optimizers_kwargs()
        self.optimize_kwargs['callback'] = self.callback

        # 2.3 Check if we need to calculate and keep additional functions
        # values in callbacks during optimization.
        self.aic_score = not self.settings.linked_snp_s
        has_boots = self.settings.directory_with_bootstrap is not None
        self.claic_score = not self.aic_score and has_boots
        self.boots = self.settings.bootstrap_data

        # 2.4 Create all necessary output dirs and files
        if settings.output_directory is not None:
            self.output_dir = os.path.join(settings.output_directory,
                                           str(self.index))

            self.resume_dir = None
            if self.settings.resume_from is not None:
                self.resume_dir = os.path.join(settings.resume_from,
                                               str(self.index))

            self.output_dir = ensure_dir_existence(self.output_dir)
            self.eval_file = os.path.join(self.output_dir, self.EVAL_FILENAME)
            self.eval_file = ensure_file_existence(self.eval_file)
            self.report_file = os.path.join(self.output_dir,
                                            self.REPORT_FILENAME)
            self.report_file = ensure_file_existence(self.report_file)
            self.save_file = os.path.join(self.output_dir, self.SAVE_FILENAME)
            self.save_file = ensure_file_existence(self.save_file)

            # Tell optimizers about output files
            self.optimize_kwargs['eval_file'] = self.eval_file
            self.optimize_kwargs['report_file'] = self.report_file
            self.optimize_kwargs['save_file'] = self.save_file
            # Create directory for pictures if needed
            if self.settings.draw_models_every_n_iteration != 0:
                self.pictures_dir = os.path.join(self.output_dir, 'pictures')
                ensure_dir_existence(self.pictures_dir)
            # Create directory for generated code if needed
            if self.settings.print_models_code_every_n_iteration != 0:
                self.code_dir = os.path.join(self.output_dir, 'code')
                ensure_dir_existence(self.code_dir)
                # If our model is built via gadma then we can write code for
                # all engines.
                if isinstance(self.model, EpochDemographicModel):
                    for engine in all_engines():
                        engine_dir = os.path.join(self.code_dir, engine.id)
                        ensure_dir_existence(engine_dir)
        # Set counters to zero for callbacks to count number of their calls
        self.draw_iter_callback_counter = 0
        self.code_iter_callback_counter = 0

        self.x_best = None
        self.y_best = None

    @property
    def model(self):
        """
        Returns current demographic model.
        """
        return self._model

    @model.setter
    def model(self, model):
        """
        Sets new demographic model and updates model in the engine.
        """
        self._model = model
        self.engine.set_model(model)

    @property
    def data(self):
        """
        Returns current data.
        """
        return self._data

    @data.setter
    def data(self, data):
        """
        Sets new data and updates data in the engine.
        """
        self._data = data
        self.engine.set_data(data)

    def base_callback(self, x, y):
        """
        Base callback:

        1) Updates values of best solution in :attr:`shared_dict`.
        2) If new best values are received then draws and generates code to\
        the `output_dir` of this run.

        :param x: Vector of values for model parameters.
        :param y: Value of log-likelihood for this values.
        """
        best_by = 'log-likelihood'
        if self.aic_score:
            n_params = self.engine.model.get_number_of_parameters(x)
            aic = get_aic_score(n_params, y)
            y_dict = {best_by: y, 'AIC score': aic}
        else:
            y_dict = y

        sign = self.global_optimizer.sign
        equal_x = False
        if self.x_best is not None:
            equal_x = list(self.x_best) == list(x)
            x_has_w = hasattr(x, 'weights')
            x_best_has_w = hasattr(self.x_best, 'weights')
            if x_has_w != x_best_has_w:
                equal_x = False
            elif x_has_w and x.metadata != self.x_best.metadata:
                equal_x = False
        if self.x_best is None or sign * self.y_best > sign * y or not equal_x:
            self.x_best = x
            self.y_best = y
            updated = self.shared_dict._put_new_model_for_process(
                self.index, best_by, (self.engine, x, y_dict))
            prefix = (self.settings.LOCAL_OUTPUT_DIR_PREFIX +
                      self.settings.LONG_NAME_2_SHORT.get(best_by, best_by))
            save_code_file = os.path.join(self.output_dir,
                                          prefix + "_model.py")
            try:
                generate_code_to_file(x, self.engine,
                                      self.settings, save_code_file)
            except Exception as e:
                pass

    def draw_iter_callback(self, x, y):
        """
        Draws best model on current iteration in the pictures directory.
        Pictures directory is located in the output directory of this run.
        Drawing happens every
        :attr:`self.settings.draw_models_every_n_iteration` iteration.

        :param x: Vector of values for model parameters.
        :param y: Value of log-likelihood for this values.
        """
        n_iter = self.draw_iter_callback_counter
        verbose = self.settings.draw_models_every_n_iteration
        if verbose != 0 and n_iter % verbose == 0:
            fig_title = f"Iteration {n_iter}, "\
                        f"Log-likelihood: {y:.2f}"
            save_file = os.path.join(self.pictures_dir,
                                     f"iteration_{n_iter}.png")
            draw_plots_to_file(x, self.engine, self.settings,
                               save_file, fig_title)
        self.draw_iter_callback_counter += 1

    def code_iter_callback(self, x, y):
        """
        Generates code of best model on current iteration to the code
        directory. Code directory is located in the output directory of this
        run. Generation happens every
        :attr:`self.settings.print_models_code_every_n_iteration` iteration.

        :param x: Vector of values for model parameters.
        :param y: Value of log-likelihood for this values.
        """
        n_iter = self.code_iter_callback_counter
        verbose = self.settings.print_models_code_every_n_iteration
        filename = f"iteration_{n_iter}.py"
        if verbose != 0 and n_iter % verbose == 0:
            if isinstance(self.model, EpochDemographicModel):
                for engine in all_engines():
                    save_file = os.path.join(self.code_dir, engine.id,
                                             filename)
                    args = self.settings.get_engine_args(engine.id)
                    engine.set_data(self.engine.data)
                    engine.data_holder = self.engine.data_holder
                    engine.set_model(self.engine.model)
                    engine.generate_code(x, save_file, *args)
            else:
                save_file = os.path.join(self.code_dir, filename)
                args = self.settings.get_engine_args()
                self.engine.generate_code(x, save_file, *args)
        self.code_iter_callback_counter += 1

    def callback(self, x, y):
        """
        Main callback for optimizers to get. It is combination of three
        callbacks:

        1) :meth:`base_callback`
        2) :meth:`draw_iter_callback`
        3) :meth:`code_iter_callback`

        :param x: Vector of values for model parameters.
        :param y: Value of log-likelihood for this values.

        """
        self.base_callback(x, y)
        try:
            self.draw_iter_callback(x, y)
        except Exception as e:
            pass
        try:
            self.code_iter_callback(x, y)
        except Exception as e:
            pass

    def intermediate_callback(self, x, y):
        """
        Almost final callback that is called after each global + local
        optimization.

        Saves AIC and CLAIC values in the :attr:`shared_dict` if needed.
        If new best by AIC or CLAIC vector then code and picture are generated.

        :param x: Vector of values for model parameters.
        :param y: Value of log-likelihood for this values.
        """
        if self.aic_score:
            best_by = "AIC score"
            n_params = self.engine.model.get_number_of_parameters(x)
            value = get_aic_score(n_params, y)
        elif self.claic_score:
            best_by = "CLAIC score"
            args = self.settings.get_engine_args()
            value = get_claic_score(self.engine, x, self.boots,
                                    args, y, return_eps=True)
        else:
            return

        y_dict = OrderedDict()
        y_dict['log-likelihood'] = y
        y_dict[best_by] = value

        if self.claic_score:
            self.shared_dict.add_model_for_process(self.index, best_by,
                                                   self.engine, x, y_dict)
        else:
            self.shared_dict.update_best_model_for_process(self.index,
                                                           best_by,
                                                           self.engine,
                                                           x, y_dict)

        # Draw and generate code for best_by model
        if self.aic_score:
            value_str = f"{value: .2f}" if value is not None else "None"
        if self.claic_score:
            if value[0] is None:
                value_str = "None"
            else:
                value_str = f"{value[0]:.2f} (eps={value[1]: .1e})"
        fig_title = f"Best by {best_by}. Log-likelihood: {y:.2f}, "\
                    f"{best_by}: {value_str}."
        prefix = (self.settings.LOCAL_OUTPUT_DIR_PREFIX +
                  self.settings.LONG_NAME_2_SHORT.get(best_by.lower(),
                                                      best_by.lower()))
        save_plot_file = os.path.join(self.output_dir, prefix + "_model.png")
        save_code_file = os.path.join(self.output_dir, prefix + "_model.py")
        try:
            draw_plots_to_file(x, self.engine, self.settings,
                               save_plot_file, fig_title)
        except Exception as e:
            print(f"{bcolors.FAIL}Run {self.index}: failed to draw model due "
                  f"to the following exception: {e}{bcolors.ENDC}")
        try:
            generate_code_to_file(x, self.engine,
                                  self.settings, save_code_file)
        except Exception as e:
            print(f"{bcolors.FAIL}Run {self.index}: failed to generate code "
                  f"due to the following exception: {e}{bcolors.ENDC}")

    def draw_model_in_output_dir(self, x, y,
                                 best_by='log-likelihood', final=True):
        """
        Draws picture of demographic model with `x` as parameters to the
        output directory of run.

        :param x: Vector of values for model parameters.
        :param y: Value of log-likelihood for this values.
        :param best_by: By what function (log-likelihood, AIC, CLAIC) this
                        colution is best.
        :param final: If True then solution is final and it will be saved by
                      `final` name.
        """
        fig_title = f"Best by {best_by} model. {best_by}: {y: .2f}"
        if final:
            prefix = self.settings.LOCAL_OUTPUT_DIR_PREFIX_FINAL
        else:
            prefix = self.settings.LOCAL_OUTPUT_DIR_PREFIX
        prefix += self.settings.LONG_NAME_2_SHORT.get(best_by, best_by)
        save_plot_file = os.path.join(self.output_dir, prefix + "_model.png")
        save_code_file = os.path.join(self.output_dir, prefix + "_model.py")
        try:
            generate_code_to_file(x, self.engine,
                                  self.settings, save_code_file)
        except Exception as e:
            pass
        try:
            draw_plots_to_file(x, self.engine, self.settings,
                               save_plot_file, fig_title)
        except Exception as e:
            pass

    def get_save_file(self):
        """
        Returns filename to save optimization run. If demographic model does
        not have structure then returns `self.save_file` else adds suffix about
        structure at the end of `self.save_file`.
        """
        if isinstance(self.model, StructureDemographicModel):
            suffix = "_".join([str(x) for x in self.model.get_structure()])
            suffix = "_" + suffix
            return self.save_file + suffix
        return self.save_file

    def run_without_increase(self, initial_kwargs={}):
        """
        Run one launch without any increase of demographic model structure.
        Runs one global+local optimization for the current model.

        :param initial_kwargs: Initial kwargs for optimization.
        """
        np.random.seed()
        self.optimize_kwargs['callback'] = self.callback
        self.optimize_kwargs['save_file'] = self.get_save_file()
        # We set some kwargs if they were not set in run_with_increase
        if self.settings.initial_structure is None:
            options = list(self.get_run_options())
            assert len(options) > 0
            restore_file, _, only_models, x_transform = options[0]
            self.optimize_kwargs['restore_file'] = restore_file
            self.optimize_kwargs['restore_points_only'] = only_models
            self.optimize_kwargs['global_x_transform'] = x_transform[0]
            self.optimize_kwargs['local_x_transform'] = x_transform[1]
        f = self.engine.evaluate
        variables = self.model.variables

        optimizer = GlobalOptimizerAndLocalOptimizer(self.global_optimizer,
                                                     self.local_optimizer)

        opt_kwargs = {**self.optimize_kwargs, **initial_kwargs}
        result = optimizer.optimize(f, variables, **opt_kwargs)
        self.intermediate_callback(result.x, result.y)
        return result

    def run_with_increase(self, initial_kwargs={}):
        """
        Run launch with increase of the demographic model structure.
        Structure of the model will be increased up to final structure. Then
        the final solution will be returned.
        Runs :meth:`run_without_increase` and ``increase_structure()`` in the
        loop.

        :param initial_kwargs: Initial kwargs for optimization.
        """
        np.random.seed()
        # Simple checks
        assert self.settings.initial_structure is not None
        assert self.settings.final_structure is not None

        options = self.get_run_options()

        restore_file, struct, only_models, x_transform = next(options)
        if struct is not None and struct != self.model.get_structure():
            warnings.warn(f"Initial structure ({init_struct}) with saved file "
                          f"of optimization from restored dir looks different "
                          f"to current ({self.settings.initial_structure}). "
                          f"It will be restored.")
            self.settings.initial_structure = struct
            self.model = self.model.from_structure(struct)
        self.optimize_kwargs['restore_file'] = restore_file
        self.optimize_kwargs['restore_points_only'] = only_models
        self.optimize_kwargs['global_x_transform'] = x_transform[0]
        self.optimize_kwargs['local_x_transform'] = x_transform[1]

        result = self.run_without_increase(initial_kwargs)
        for restore_file, structure, only_models, x_transform in options:
            X_init = result.X_out
            if self.model.get_structure() != structure:
                self.model, X_init = self.model.increase_structure(structure,
                                                                   X=X_init)
            Y_init = copy.copy(result.Y_out)
            self.optimize_kwargs['X_init'] = X_init
            self.optimize_kwargs['Y_init'] = Y_init
            self.optimize_kwargs['restore_file'] = restore_file
            self.optimize_kwargs['restore_points_only'] = only_models
            self.optimize_kwargs['global_x_transform'] = x_transform[0]
            self.optimize_kwargs['local_x_transform'] = x_transform[1]
            result = self.run_without_increase()
#        self.final_callback()
        return result

    def get_run_options(self):
        """
        Returns iterator of run options, each element has four options:

        1) File to restore optimization from. If None then no resume.
        2) Structure of the demographic model for the run.
        3) Bool `points_only` - if True then resumed run uses old points as\
        initial points only and optimization is run from the beginning.
        4) Function of x transformation - for case when restored vectors\
        should be transformed somehow before they will be used in the\
        optimization.

        """
        restore_files = []
        only_models = []
        structures = []
        if self.resume_dir is not None:
            for filename in os.listdir(self.resume_dir):
                file_path = os.path.join(self.resume_dir, filename)
                if filename.startswith(self.SAVE_FILENAME):
                    if self.settings.initial_structure is None:
                        if filename == self.SAVE_FILENAME:
                            restore_files.append(file_path)
                            break
                        continue
                    if len(filename) == len(self.SAVE_FILENAME):
                        warnings.warn(f"File {file_path} has name like saved"
                                      f" file of optimization but has no "
                                      f"structure at the end of the name. "
                                      f"So it is ignored.")
                        continue
                    restore_files.append(file_path)
                    strct_str = filename[len(self.SAVE_FILENAME)+1:]
                    structures.append([int(x) for x in strct_str.split('_')])
            if self.settings.initial_structure is not None:
                restore_files, structures = sort_by_other_list(
                    restore_files, structures, key=lambda x: sum(x))
            else:
                x_transform = (None, None)
                if self.settings.generate_x_transform:
                    x_transform = (ident_transform, ident_transform)
                return iter([(restore_files[0], None,
                              self.settings.only_models, x_transform)])

            some_file_not_valid = False
            for i in range(len(restore_files)):
                if some_file_not_valid:
                    restore_files[i] = None
                    only_models.append(False)
                    continue
                gs = self.global_optimizer.valid_restore_file(restore_files[i])
                ls = self.local_optimizer.valid_restore_file(restore_files[i])
                only_models.append(False)
                if not gs or not ls:
                    some_file_not_valid = True
            if self.settings.only_models:
                only_models.append(True)
                structures.append(structures[-1])

        if self.settings.initial_structure is None:
            return iter([(None, None, False, (None, None))])

        final_sum = sum(self.settings.final_structure)
        initial_sum = sum(self.settings.initial_structure)
        res_files, res_strct, res_bools, res_trans = [], [], [], []
        if self.settings.generate_x_transform:
            old_params = os.path.join(self.settings.resume_from, 'params_file')
            old_extra = os.path.join(self.settings.resume_from,
                                     'extra_params_file')
            old_settings = SettingsStorage.from_file(old_params, old_extra)
            old_init_model = old_settings.get_model()
        for i in range(final_sum - initial_sum + 1 +
                       int(self.settings.only_models)):
            if i >= len(restore_files):
                restore_file = None
            else:
                restore_file = restore_files[i]
            if i >= len(structures):
                structure = None
            else:
                structure = structures[i]
            if i >= len(only_models):
                restore_points_only = False
            else:
                restore_points_only = only_models[i]
            if not self.settings.generate_x_transform or restore_file is None:
                gs_x_transform = None
                ls_x_transform = None
            else:
                old_init_model = old_init_model.from_structure(structure)
                gs_x_transform = partial(
                    self.model.transform_values_from_other_model,
                    copy.deepcopy(old_init_model))
                copy_old_model = copy.deepcopy(old_init_model)
                copy_self_model = copy.deepcopy(self.model)
                copy_old_model.have_dyns = False
                copy_old_model = copy_old_model.from_structure(structure)
                copy_self_model.have_dyns = False
                copy_self_model = copy_self_model.from_structure(structure)
                ls_x_transform = partial(
                    copy_self_model.transform_values_from_other_model,
                    copy_old_model)

            res_files.append(restore_file)
            res_strct.append(structure)
            res_bools.append(restore_points_only)
            res_trans.append((gs_x_transform, ls_x_transform))
        return zip(res_files, res_strct, res_bools, res_trans)

    def run(self, initial_kwargs={}):
        """
        Main method of the class to run optimization. Runs
        :meth:`run_without_increase` or :meth:`run_with_increase` according to
        `settings`.
        """
        print(f'Run launch number {self.index}\n', end='')
        if self.index in self.shared_dict.dict:
            del self.shared_dict.dict[self.index]
        if self.settings.initial_structure is None:
            result = self.run_without_increase(initial_kwargs)
        else:
            result = self.run_with_increase(initial_kwargs)
        # draw final model in output dir
        self.draw_model_in_output_dir(result.x, result.y)
        print(f'Finish genetic algorithm number {self.index}\n', end='')
        return result



#!/usr/bin/env python3

############################################################################
# Copyright (c) 2018 Noskova Ekaterina
# All Rights Reserved
# See the LICENSE file for details
############################################################################

try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages


import os, sys


NAME = 'gadma'

VERSION = '2.0.0rc1'
SUPPORTED_PYTHON_VERSIONS = ['3.6', '3.7']


# Check python version
if sys.version[0:3] not in SUPPORTED_PYTHON_VERSIONS:
    sys.stderr.write("Python version " + sys.version[0:3] + " is not supported!\n" +
          "Supported versions are " + ", ".join(SUPPORTED_PYTHON_VERSIONS) + "\n")
    sys.stderr.flush()
    sys.exit(1)


# Create a simple version.py module; less trouble than hard-coding the version
with open(os.path.join('gadma', 'version.py'), 'w') as f:
    f.write('__version__ = %r\nversion = __version__\n' % VERSION)
    f.write('\n# This is a new line that ends the file.\n')


# Load up the description from README.rst
with open('README.md') as f:
    DESCRIPTION = f.read()

requirements = ['numpy', 'scipy', 'matplotlib',
                'Pillow', 'Cython', 'mpmath', 'nlopt', 'ruamel.yaml',
                'dadi']

setup(
    name=NAME,
    version=VERSION,
    author='Ekaterina Noskova',
    author_email='ekaterina.e.noskova@gmail.com',
    url='https://github.com/ctlab/GADMA',
    description='Genetic Algorithm for Demographic Inference',
    long_description=DESCRIPTION,
    long_description_content_type='text/markdown',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development',
    ],
    packages=find_packages(exclude=['examples', 'tests']),
    include_package_data=True,
    package_data={
        'gadma.cli': ['*.py',  'params_template', 'extra_params_template', 'test_settings']
    },
    data_files=[('fs_examples', [os.path.join('fs_examples', 'test.fs')]), ("", ["LICENSE"])],
    install_requires=requirements,
    entry_points={
        'console_scripts': ['gadma = gadma.core:main',
            'gadma-run_ls_on_boot_data = gadma.run_ls_on_boot_data:main',
            'gadma-get_confidence_intervals = gadma.get_confidence_intervals:main']
    },
)



from ..utils import PopulationSizeVariable, TimeVariable, MigrationVariable,\
                    DynamicVariable, FractionVariable

# Main options. Output and input.
output_directory = None
input_file = None
# input_data = None
# number_of_populations = None
population_labels = None
projections = None
outgroup = None
sequence_length = None
linked_snp_s = True
directory_with_bootstrap = None
# boots = None

# Pipeline
theta0 = None
time_for_generation = None
# multinom = None
only_sudden = False
pts = None
engine = 'moments'
relative_parameters = False
no_migrations = False
symmetric_migrations = False
split_fractions = True

# Custom model
custom_filename = None
model_func = None
lower_bound = None
upper_bound = None
parameter_identifiers = None

# Structure of models
initial_structure_unit = 1
initial_structure = None
final_structure = None

# Time bounds
upper_bound_of_first_split = None
upper_bound_of_second_split = None

# GA options
size_of_generation = 10

fractions = [0.2, 0.3, 0.3]
n_elitism = 2
p_mutation = 0.3
p_crossover = 0.3
p_random = 0.2

mean_mutation_strength = 0.2
const_for_mutation_strength = 1.01

mean_mutation_rate = 0.2
const_for_mutation_rate = 1.02

stuck_generation_number = 100
eps = 1e-2

# just for logging evaluations
# output_log_file = None
# max_num_of_eval = None # maximum number of logll eval.
# num_init_pts = None # can get value from Inference.optimize_ga

# Local search
local_optimizer = 'BFGS_log'

# Printing and drawing
print_models_code_every_n_iteration = 0
silence = False
draw_models_every_n_iteration = 0
units_of_time_in_drawing = 'generations'
const_of_time_in_drawing = 1.0
vmin = 1

number_of_repeats = 1
number_of_processes = 1
test = False
resume_from = None
only_models = False
generate_x_transform = False

# Extra parameters

# Bounds on models parameters. They are relative to N_A (!)
min_n = PopulationSizeVariable.default_domain[0]
max_n = PopulationSizeVariable.default_domain[1]
min_t = TimeVariable.default_domain[0]
max_t = TimeVariable.default_domain[1]
min_m = MigrationVariable.default_domain[0]
max_m = MigrationVariable.default_domain[1]

# Parameters for local search alg
# ls_verbose = None
# ls_flush_delay = 0.5
# ls_epsilon = 1e-3
# ls_gtol = 1e-05
# ls_maxiter = None
# for hill climbing
# hc_mutation_rate = None
# hc_const_for_mutation_rate = None
# hc_stop_iter = None

# Options of mutation, crossing and random generating
random_n_a = True
# multinom_cross = False
# multinom_mutate = False

# Options of printing summary information about repeats
time_to_print_summary = 1  # min

verbose = 1

# Options of distributions
# distribution = 'normal'  # can be 'uniform'
# std = None  # std for normal dist

# Some options about drawing plots:
# matplotlib_available = False
# pil_available = False
# moments_available = False

X_init = None
Y_init = None
mutation_rate = None

# Additional constants
P_IDS = {'n': PopulationSizeVariable, 't': TimeVariable,
         'm': MigrationVariable, 'd': DynamicVariable, 's': FractionVariable,
         'f': FractionVariable}
LONG_NAME_2_SHORT = {"log-likelihood": "logLL",
                     "aic score": "aic",
                     "claic score": "claic"}
BASE_OUTPUT_DIR_PREFIX = "best_"
BASE_OUTPUT_DIR_PREFIX_FINAL = "best_"
LOCAL_OUTPUT_DIR_PREFIX = "current_best_"
LOCAL_OUTPUT_DIR_PREFIX_FINAL = "final_best_"
